__version__ = '0.5.9'

from . import DE, GA, PSO, SA, ACA, AFSA, IA


def start():
    print('''
    scikit-opt import successfully,
    version: {version}
    Author: Guo Fei,
    Email: guofei9987@foxmail.com
    repo: https://github.com/guofei9987/scikit-opt,
    documents: https://scikit-opt.github.io/
    '''.format(version=__version__))



import datetime
import json
import os
import sys

import docx
import numpy as np
import pandas as pd
import win32com.client
from yangke.common.QtImporter import (QFont, QCheckBox, QApplication, QPushButton, QFileDialog, QProgressBar,
                                      QDialog, QVBoxLayout, QLabel, QButtonGroup, QHBoxLayout, QRadioButton,
                                      QDialogButtonBox, QMessageBox, QtCore, QObject, pyqtSignal
                                      )

from yangke.base import get_temp_para, save_temp_para, yield_all_file, \
    start_threads, get_datetime_col_of_dataframe
from yangke.common.config import logger
from yangke.performance.basic_para import pressure_absolute, pressure_relative  # 不能删除，本文件有隐含式引用
from yangke.common.qt import YkWindow, YkDataTableWidget, set_menu_bar
from yangke.core import str_in_list
from yangke.common.fileOperate import read_csv_ex, write_as_pickle, read_from_pickle
from yangke.base import merge_two_dataframes, save_as_xlsx
import time

from yangke.performance.dp.const_name import suffix_p_10, suffix_dcs_10, suffix_imp_20, suffix_dcs_20, suffix_p_20, \
    suffix_imp_10
from yangke.performance.dp.cutting_period import pre_deal_and_cutting_period

warning_info = {"imp": {}, "p": {}, "dcs": {}}  # 记录所有数据处理的警告信息，最后写入汇总文件


def func1():
    xl = win32com.client.Dispatch('Excel.Application')
    xl.Visible = True
    xla_file = r"D:\Users\2020\性能试验\陈会勇\数据处理\ReadTestData（最新）.xla"
    xl.Workbooks.Open(xla_file)

    # data_file = r"D:\Users\2020\性能试验\陈会勇\数据处理\试验数据\IMP数据\3VWO-背压16.8.dat"
    # new_data_file = os.path.join(os.path.dirname(data_file), os.path.basename(data_file).split(".")[0] + ".xlsx"
    # xl.Workbooks.Open(data_file)
    # xl.Application.Save()
    # xl.Application.Quit()
    # print("end")


def add_atmosphere(data_series: pd.Series):
    """
    对数据进行大气压修正，将修正的测点的单位改为MPa
    :param data_series: IMP或WSN试验采集仪采集的试验测点，data_series.index是试验测点名称，data_series.values是试验测点数值
    :return:
    """

    def get_p_env_idx(test_point_list: list) -> str:
        """
        找出参数列表中大气压对应的索引

        :param test_point_list:
        :return: 返回大气压对应的变量的名称，是一个str
        """
        for idx in test_point_list:
            name_str = idx
            if "大气压" in name_str:
                return idx
        logger.error("未找到大气压数据")
        raise ValueError("未找到大气压数据")

    p_env_idx = get_p_env_idx(data_series.index)
    p_env = data_series[p_env_idx]
    for name, value in data_series.items():
        name = str(name)
        if "pa" not in name.lower():  # 说明不是压力数据
            continue
        if "流量" in name or "差压" in name or "PD" in name:
            continue
        # 有限依据表号判断

        if str_in_list(name, pressure_relative, revert=True):
            data_series[name] = (value + p_env) / 1000
            data_series.rename(index={name: name.replace("kPa", "MPa")}, inplace=True)  # 更改单位
        elif str_in_list(name, pressure_absolute, revert=True):
            continue
        else:
            logger.warning(f"无法判断压力数据是绝压还是表压，测点名称为：{name}")
    return data_series


def get_test_time(folder):
    """
    获取各工况的数据采集开始和结束时间

    :param folder:
    :return:
    """
    condition_time = {}
    name_idx = -1
    date_idx = -1
    time_idx = -1
    file = None
    for file in yield_all_file(folder, [".docx"]):
        if not str_in_list(os.path.splitext(os.path.basename(file))[0],
                           ["工况记录单", "工况确认时间", "试验工况时间", "工况确认单"],
                           revert=True):
            continue
        logger.debug("读取工况记录单数据...")
        doc = docx.Document(file)
        for table in doc.tables:
            row = table.rows[0]  # 根据表格第一行标题确定每列的数据是什么
            for col_idx, cell in enumerate(row.cells):
                if "名称" in cell.text.strip() or cell.text.strip() == "工况" or cell.text.strip() == "试验工况":
                    name_idx = col_idx
                elif cell.text.strip() == "日期":
                    date_idx = col_idx
                elif cell.text.strip() == "时间" or cell.text.strip() == "记录数据时间":
                    time_idx = col_idx
            if name_idx == -1:  # 说明按上述标准没有找到工况名称列，进一步检索表头"工况"
                for col_idx, cell in enumerate(row.cells):
                    if "工况" in cell.text.strip() and col_idx != date_idx and col_idx != time_idx:
                        name_idx = col_idx
                        break

            for row_idx, row in enumerate(table.rows[1:]):
                try:
                    cells = row.cells
                    condition_name = cells[name_idx].text.strip()
                    if condition_name == "":
                        continue
                    date_str = cells[date_idx].text.strip()
                    time_str = cells[time_idx].text.strip().replace("：", ":")
                    time_str = time_str.replace(" ", "")
                    if time_str == "":
                        logger.error(f"工况记录单{file}中存在数据不全的工况->工况名为：{condition_name}")
                        sys.exit(0)
                    date_str = str(pd.Timestamp(date_str)).split(" ")[0]  # 使用pd.Timestamp接受多种格式的字符串，返回%Y-%m-%d
                    start_time_str, end_time_str = time_str.split("-")
                    start_time = time.strptime(start_time_str, "%H:%M")
                    end_time = time.strptime(end_time_str, "%H:%M")
                    start_datetime_str = date_str + " " + time.strftime("%H:%M", start_time)
                    if end_time < start_time:  # 一般单工况持续时间小于3小时，如果结束时间比开始时间小，说明跨了一天，日期需要加1
                        date = datetime.datetime.strptime(date_str, "%Y-%m-%d") \
                               + datetime.timedelta(days=1)
                        end_datetime_str = date.strftime("%Y-%m-%d") + " " + time.strftime("%H:%M", end_time)
                    else:
                        end_datetime_str = date_str + " " + time.strftime("%H:%M", end_time)
                    condition_time.update(
                        {condition_name: {"start_time": start_datetime_str, "end_time": end_datetime_str}})
                    logger.debug(f"{condition_name} -> {json.dumps(condition_time.get(condition_name))}")
                except:  #
                    break  # 处理表格中某一行出错时，退出行遍历
            break  # 只便利word中的第一个table，第一个table出力完成后推出tables便利
        break  # 只处理检索到的第一个word文档，处理后则退出文件遍历
    return condition_time, file


def delete_error_data(df, bound=0.3, method="relative_error"):
    """
    删除数据集中偏差过大的数据，df数据的columns是变量名，index是数字索引或时间索引，values全是变量值

    :param df:
    :param bound:
    :param method: 剔除错误数据的方法，relative_error—根据相对误差剔除跳数，trend—根据数据趋势剔除跳数
    :return:
    """
    time_col = get_datetime_col_of_dataframe(df)
    if time_col != -1:
        data_no_time = df.set_index("DateTime")
    else:
        data_no_time = time_col
    name_series = data_no_time.columns
    mean_v = data_no_time.mean()
    min_v = data_no_time.min()
    max_v = data_no_time.max()
    criteria = (max_v - min_v) / mean_v > bound  # 选出最大值最小值差超过平均50%的Series
    data_no_time = data_no_time.T
    need_modified_series = data_no_time[criteria]
    need_modified_series_idx = need_modified_series.index
    for idx in need_modified_series_idx:
        current_series: pd.Series = need_modified_series.loc[idx, :]
        if "减温水" in idx or "水位" in idx or "液位" in idx:  # 减温水的数据不判断，因为减温水会间歇运行，其数据就是波动的
            # 将修改后的series替换到元数据dataframe中
            continue
        max_s = current_series.max()
        min_s = current_series.min()
        mean_s = current_series.mean()
        idx_deleted = []
        while (max_s - mean_s) / mean_s > bound or (mean_s - min_s) / mean_s > bound:
            # 找出参数行中需要删除的值，并将其置为np.nan，则后续求mean、max、min等参数时会忽略np.nan值
            low_bound, high_bound = mean_s * (1 - bound), mean_s * (1 + bound)
            # noinspection All
            del_idx = current_series[(current_series < low_bound) | (current_series > high_bound)].index
            idx_deleted.extend(list(del_idx))
            current_series[del_idx] = np.nan
            # 再次计算该数据列最大最小值的差别是否超出平均数的bound之外
            max_s = current_series.max()
            min_s = current_series.min()
            mean_s = current_series.mean()

        logger.info(f"剔除的数据索引 第{idx}列， 第{str(idx_deleted)}行")  # 这里提示的相当于原数据的行列
        data_temp = data_no_time.copy()
        data_temp.loc[idx] = current_series  # 将修改后的series替换到元数据dataframe中
        data_no_time = data_temp
    data_no_time = data_no_time.T
    df = data_no_time.reset_index()
    return df


def deal_liquid_level(data: pd.DataFrame, **kwargs):
    """
    处理数据中的液位测点，将测点名改为 ”*水位下降_*"，例如原测点名称为"除氧器水位_mm"，则处理后的测点名成为"除氧器水位下降_mm"，处理后data中
    液位列的所有数据都是下降的数值。

    2021.03.20添加功能：在最终结果中输出水位数据的前{n}个与后{n}个的平均值

    默认inplace=True，且不可更改

    :param data: 测点数据
    :return:
    """
    num = int(kwargs.get("水位计算点数"))
    data.sort_values(by="DateTime", inplace=True)
    for col_name in data.columns:
        if ("水位" in col_name or "液位" in col_name) and ("平均值" not in col_name and "下降值" not in col_name):
            name_new = col_name + "_下降值"
            header_mean = data[col_name].head(num).mean()
            tail_mean = data[col_name].tail(num).mean()
            lvl_drop = header_mean - tail_mean
            if lvl_drop < 0:
                condition = kwargs.get("condition")
                logger.warning(f"{condition}试验期间{col_name}上升，存在不明流量或泄露阀门")
                insert_warning("imp", "水位警告", f"{condition}试验期间，（{col_name}）上升，存在不明流量或泄露阀门")
            data[col_name] = lvl_drop
            data.rename(columns={col_name: name_new}, inplace=True)
            cols_ = data.columns.get_loc(name_new)
            data.insert(cols_, f"{col_name}_前{num}个平均值", header_mean)
            data.insert(cols_ + 1, f"{col_name}_后{num}个平均值", tail_mean)
    return data


def insert_warning(main_type, sub_type, content):
    """
    插入警告信息

    :param main_type: 警告的主类别，只可能是 "imp", "p", "dcs"
    :param sub_type: 警告的子类别，可自由定义
    :param content: 警告信息内容
    :return:
    """
    warn_dict: dict = warning_info.get(main_type)
    if warn_dict is None:
        logger.warning("警告主类别不存在")
        return
    temp = warn_dict.get(sub_type)
    if temp is None:
        warn_dict.update({sub_type: [content]})
    else:
        temp.append(content)
        warn_dict.update({sub_type: temp})


class ConfirmDialog(QDialog):
    """
    用于确认压力测点是表压还是绝压的弹出式对话框
    """

    def __init__(self, parent=None, var_list=None, abs_list=None, rel_list=None):
        super(ConfirmDialog, self).__init__(parent=parent)
        if var_list is None:
            var_list = []
        self.var_list = var_list
        self.abs_list = abs_list
        self.rel_list = rel_list
        self.button_group = []
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle("表压/绝压确认输入窗口")
        self.setGeometry(400, 400, 500, 200)

        v_box = QVBoxLayout()
        for i, var in enumerate(self.var_list):
            tag = str(var).split("_")[2]
            btn_group = QButtonGroup()
            h_box = QHBoxLayout()
            label = QLabel(str(var))
            label.setFixedWidth(300)
            radio_btn_y = QRadioButton('绝压')
            radio_btn_n = QRadioButton('表压')
            if self.abs_list is None and self.rel_list is None:
                if "pg" in tag.lower():
                    radio_btn_n.setChecked(True)
                else:  # PD或PA，差压或绝压
                    radio_btn_y.setChecked(True)
            elif self.abs_list is not None and self.rel_list is not None:
                if var in self.abs_list:
                    radio_btn_y.setChecked(True)
                else:
                    radio_btn_n.setChecked(True)
            btn_group.addButton(radio_btn_y)
            btn_group.addButton(radio_btn_n)
            self.button_group.append(btn_group)
            h_box.addWidget(label)
            h_box.addWidget(radio_btn_y)
            h_box.addWidget(radio_btn_n)
            v_box.addLayout(h_box)

        buttons = QDialogButtonBox(QDialogButtonBox.Ok)

        v_box.addWidget(buttons)
        # noinspection all
        buttons.accepted.connect(self.accept)

        self.setLayout(v_box)

    def get_data(self):  # 定义用户获取数据的方法
        value = []
        for i, var in enumerate(self.var_list):
            value.append(self.button_group[i].checkedButton().text())
        return self.var_list, value


def deal_poor_or_bad_parameters(poor_series: pd.Series):
    """
    处理DCS导出的质量较差的数据。
    DCS数据以B结尾，表示该数据是错误的。
    DCS数据以P结尾，表示该数据质量较差

    :param poor_series:
    :return:
    """
    values = []
    try:
        for value in poor_series:
            value = str(value).replace("B", "")  # DCS数据以B结尾，表示该数据是错误的
            value = value.replace("P", "")  # DCS数据以P结尾，表示该数据质量较差
            if value.strip() == "":
                continue
            values.append(float(value))
        return pd.Series(values).mean()
    except ValueError:
        return None


class MainWindow(YkWindow, QObject):

    def __init__(self):
        # noinspection PyTypeChecker
        self.table_widget: YkDataTableWidget = None
        self.data_folder = None
        self.project_file = None
        self.test_time_file = None  # 工况记录单文件
        self.test_time = {}  # 工况记录单中对应的工况时间信息
        self.bound = 0.3
        self.progressBar = QProgressBar()
        self.config = {}  # 配置项
        super(MainWindow, self).__init__()
        self.imp_file = None
        self.dcs_file = None
        self.power_file = None
        self.unknown_file = None  # 无法识别格式的文件中数据的汇总文件
        self.dropped_dcs_cols = []  # dcs数据处理过程总丢弃的非数值型数据列
        self.point_description = None  # 和利时DCS的点表文件，记录了KKS对应的点名，用于更新导出的数据文件中的KKS码
        self.temp_para = None  # 用于子线程和主线程传递数据时使用
        self.project = {}  # 项目文件，保存时保存该参数
        # noinspection all
        self.yk_signal.connect(self.on_message_changed)
        set_menu_bar(self, from_file=os.path.join(os.path.dirname(__file__), "ui/ui_menu_dp.yaml"))  # 设置菜单栏
        self.resize(1200, 800)
        self.center()
        self.setWindowTitle("性能试验数据处理v2.0")

    def gather_imp_or_wsn(self, files: list):
        """
        汇总imp或wsn数据到一个文件中，这里files读入的数据格式是固定的

        :param files:
        :return:
        """
        if len(files) == 0:
            return
        conditions = []
        df = None
        tmp_file = None
        data_list = []
        for file in files:
            data = read_csv_ex(file)
            condition = os.path.basename(file).split("_")[0]
            data = delete_error_data(data, bound=0.3)  # 清洗数据
            kwargs_ = {"condition": condition}
            kwargs_.update(self.config)
            data = deal_liquid_level(data, **kwargs_)
            data_mean = data.mean(numeric_only=True)  # 拿到评测点平均值
            self.temp_para = data_mean
            # noinspection all
            # self.yk_signal.emit("eval", "self.ask_for_absolute_or_relative()")
            self.ask_for_absolute_or_relative()
            while self.temp_para is not None:  # 当self.temp_para is None时，表示信号调用的新线程处理完毕
                time.sleep(0.5)
            try:
                data_series = add_atmosphere(data_mean)  # 添加大气压修正，且将单位修正带表压MPa，绝压kPa
            except:  # 有些特殊试验，现场采集数据没有大气压数据，则不进行大气压力修正
                data_series = data_mean
            data_series.name = condition
            data_list.append(data_series)

        df = pd.concat(data_list, axis=1)
        df.dropna(how="all", inplace=True)
        df = self.sort_conditions_add_time_section(df)
        outfile = self.imp_file
        try:
            df.to_excel(outfile, sheet_name="大气压修正汇总")
        except PermissionError:
            logger.error(f"文件被占用，无法写入：{outfile}")
        self.table_widget.display_dataframe(df, row_index=3, col_index=0, digits=4)

        self.table_widget.setColumnWidth(0, 200)  # 第一列是参数名称，将其宽度调宽一点

        return outfile

    def gather_power(self, files: list):
        """
        汇总所有工况的功率数据到一个excel中，有些机组安装了一个以上的功率表，需要特殊处理

        :param files:
        :return:
        """
        if len(files) == 0:
            return
        series = []
        names = []
        for file in files:
            data = read_csv_ex(file, index_col=0)
            data_no_time = data.iloc[:, 1:].copy()
            mean = data_no_time.mean(axis=0, numeric_only=True)
            condition_name = os.path.splitext(os.path.basename(file))[0].replace(suffix_p_10, "")
            # mean.NAME = condition_name
            series.append(mean)
            names.append(condition_name)
        df = pd.concat(series, axis=1)  # 设置names无效
        df = df.set_axis(names, axis=1)  # 显式设置列标题
        # 将df中工况按工况记录单排序
        df = self.sort_conditions_add_time_section(df)
        outfile = self.power_file
        try:
            df.to_excel(outfile)
        except PermissionError:
            logger.error(f"无写入文件的权限：{outfile}，请检查文件是否占用！")
        return outfile

    def gather_dcs(self, files: list):
        if len(files) == 0:
            return
        series = []
        for file in files:
            data = read_csv_ex(file, index_col=0)
            data_no_time = data
            condition = os.path.basename(file).split("_")[0]
            kwargs_ = {"condition": condition}
            kwargs_.update(self.config)
            data_no_time = deal_liquid_level(data_no_time, **kwargs_)
            mean: pd.Series = data_no_time.mean(axis=0)
            if len(mean) < data_no_time.shape[1]:
                # 说明有列在求平均后丢失了
                # 找到丢失的列名
                columns_origin = set(data_no_time.columns)
                columns_mean = set(mean.index)
                columns_dropped = columns_origin - columns_mean
                for column in columns_dropped:
                    temp = deal_poor_or_bad_parameters(data_no_time[column])
                    if temp is not None:
                        mean[column + "_poor"] = temp
            condition_name = os.path.splitext(os.path.basename(file))[0].replace(suffix_dcs_10, "")
            mean.name = condition_name
            mean.sort_index(inplace=True)
            series.append(mean)
        df = pd.concat(series, axis=1)
        if self.point_description is not None:  # 表明存在点表文件，需要根据点表文件更新DCS导数中的KKS码
            def split(text: str):
                return text.split(".")[0]

            df.set_index(df.index.map(split), inplace=True)  # 删除导出的点名后表示值计算方法的后缀，如.AV：平均值等
            df.rename(index=self.point_description, inplace=True)  # 将KKS码替换为点描述

        df = self.sort_conditions_add_time_section(df)
        outfile = self.dcs_file
        df.to_excel(outfile)
        return outfile

    def gather_unknown(self, files: list):
        """
        汇总未识别格式的文件的内容

        :param files:
        :return:
        """
        if len(files) == 0:
            return
        series = []
        for file in files:
            data = read_csv_ex(file, index_col=0)
            # 删掉标题带有序号的列
            filter_cols = [col for col in list(data.columns) if "序号" not in col]
            data_no_time = data[filter_cols].copy()
            condition = os.path.basename(file).split("_")[0]
            kwargs_ = {"condition": condition}
            kwargs_.update(self.config)
            data_no_time = deal_liquid_level(data_no_time, **kwargs_)
            mean = data_no_time.mean(axis=0)
            mean.NAME = condition
            series.append(mean)
        df = pd.concat(series, axis=1)
        outfile = self.unknown_file
        df = self.sort_conditions_add_time_section(df)
        df.to_excel(outfile)
        return outfile

    def sort_conditions_add_time_section(self, df: pd.DataFrame):
        """
        按照工况记录单对结果文件中的列进行排序，在数据区域前三行添加每个工况的起止时间

        :param df:
        :return:
        """
        missing_cols = set(self.test_time.keys()) - set(df.columns)
        for c in missing_cols:  # 如果数据处理结果中缺少某个工况的数据，则用np.nan代替
            df[c] = np.nan
        df = df[list(self.test_time.keys())]
        start_date = []
        start_time = []
        end_time = []
        for k, v in self.test_time.items():
            start_date.append(v['start_time'][:10])
            start_time.append(v['start_time'][11:])
            end_time.append(v['end_time'][11:])
        df = df.T
        df.insert(0, '结束时间', end_time)
        df.insert(0, '开始时间', start_time)
        df.insert(0, '开始日期', start_date)
        df = df.T
        return df

    def choose_file(self):
        table = self.table_widget
        folder = QFileDialog.getExistingDirectory(parent=self, caption="选择数据存储目录",
                                                  directory=self.data_folder)
        if folder != self.data_folder:
            self.data_folder = folder
            save_temp_para("directory", folder)
        table.set_value("目录", folder)

    def deal(self):
        table = self.table_widget
        self.progressBar.setVisible(True)
        self.statusBar().showMessage("处理试验数据...")
        self.progressBar.setValue(0)
        self.config.update({"水位计算点数": self.table_widget.get_value("水位计算点数"),
                            "使用表号确定压力类型": self.table_widget.get_value("使用表号确定压力类型"),
                            "未知类型文件区分文件名": self.table_widget.get_value("未知类型文件区分文件名")})

        self.table_widget.get_button("自动处理").setEnabled(False)
        # start_threads(targets=task_func)
        self.task_func(table)

    def init_ui(self):
        self.data_folder = get_temp_para('directory', get_func_or_default_value=lambda: "c:\\")
        self.project_file = os.path.join(self.data_folder, "project.dat")

        self.config.update(read_from_pickle(self.project_file) or {})

        # YkDataTableWidget.calculate = calculate
        self.table_widget = YkDataTableWidget(from_file=os.path.join(os.path.dirname(__file__), "ui/table_data1.yaml"),
                                              root_window=self)
        for k, v in self.config.items():  # 按项目配置文件设置表格属性
            self.table_widget.set_value(k, v)
        self.setCentralWidget(self.table_widget)
        if self.data_folder is not None and self.data_folder != "c:\\":
            self.table_widget.set_value("目录", self.data_folder)

        self.statusBar().addPermanentWidget(self.progressBar)
        self.progressBar.setGeometry(0, 0, 50, 2)
        self.progressBar.setValue(0)
        self.progressBar.setVisible(False)

    def task_func(self, table):
        """
        启动新线程运行耗时较长的方法
        不能在该线程中创建或修改父线程的GUI元素，如果需要修改GUI，必须使用yk_signal触发QEvent事件调用父线程的方法
        :return:
        """
        folder = table.get_value("目录")
        self.imp_file = os.path.join(folder, suffix_imp_20 + ".xlsx")
        self.dcs_file = os.path.join(folder, suffix_dcs_20 + ".xlsx")
        self.power_file = os.path.join(folder, suffix_p_20 + ".xlsx")
        self.unknown_file = os.path.join(folder, "unk汇总_修正.xlsx")  # 未识别格式的文件的统计数据
        self.test_time, self.test_time_file = get_test_time(folder)  # 查询工况记录单
        logger.debug(f"工况记录单文件为：{self.test_time_file}")
        number_con = len(self.test_time.keys())
        logger.info(f"工况记录单中查寻到的工况共{number_con}个！")
        logger.info(str(list(self.test_time.keys())))
        dcs_tested = []
        dcs_individual_file = []  # dcs预处理后的单个储存文件
        imp_or_wsn_tested = []
        imp_individual_file = []  # imp数据预处理后的单个储存文件列表
        power_tested = []
        power_individual_file = []  # 功率数据预处理后的单个储存文件列表
        unknown_tested = []
        unknown_individual_file = []
        step = 0
        for file in yield_all_file(folder, [".csv", ".xlsx", ".dat", ".docx", ".txt", ".xls"]):
            step = step + 1 if step < 100 else 0
            # self.progressBar.setValue(step)
            # noinspection all
            self.progressBar.setValue(step)
            QApplication.processEvents()
            # self.yk_signal.emit("eval", "self.progressBar.setValue(" + str(step) + ")")
            if os.path.splitext(os.path.basename(file))[0].endswith("修正") or "ignore" in os.path.abspath(
                    file) or "忽略该文件" in os.path.abspath(file):
                logger.info(f"忽略文件{file}")
                continue
            logger.debug(f"处理文件 {file}...")
            if os.path.splitext(file)[1] == ".xls":
                # 将xls格式文件另存为csv文件，csv的兼容性最好，然后进行进一步的处理
                file = save_as_xlsx(file, engine="WPS", ext="xlsx")
            # 预处理文件，如果文件中包含“时间-数据”条目，则剪切到对应工况的时间段，返回文件类型及剪切后的文件
            cutting_result = pre_deal_and_cutting_period(file, self.test_time, self.bound, self.config)
            if cutting_result:
                if isinstance(cutting_result, tuple) and cutting_result[0] == "power_slice":
                    power_tested.extend(cutting_result[2])
                    power_individual_file.extend(cutting_result[1])
                elif isinstance(cutting_result, tuple) and cutting_result[0] == "imp_slice":
                    imp_or_wsn_tested.extend(cutting_result[2])
                    imp_individual_file.extend(cutting_result[1])
                    self.config.update({"imp_or_wsn_tested": imp_or_wsn_tested})
                elif isinstance(cutting_result, tuple) and cutting_result[0] == "wsn_slice":
                    for condition in cutting_result[2]:
                        file1 = cutting_result[1][
                            cutting_result[2].index(condition)]  # 找到condition的索引，然后按该索引拿到对应的文件名
                        if condition in imp_or_wsn_tested:
                            file2 = imp_individual_file[imp_or_wsn_tested.index(condition)]
                            data1 = pd.read_csv(file1)
                            data2: pd.DataFrame = pd.read_csv(file2)
                            data2, dropped_cols = merge_two_dataframes(data1, data2)
                            data2[data2 == 0.000] = np.nan
                            data2.to_csv(file2, index=False)
                            os.remove(file1)
                        else:
                            file2 = os.path.join(os.path.dirname(file1), f"{condition}{suffix_imp_10}.csv")
                            if os.path.exists(file2):
                                os.remove(file2)
                            os.rename(file1, file2)
                            imp_or_wsn_tested.append(condition)
                            imp_individual_file.append(file2)
                elif isinstance(cutting_result, tuple) and cutting_result[0] == "dcs_slice":
                    # cutting_result[1] 为DCS处理后文件
                    for condition in cutting_result[2]:
                        file1 = cutting_result[1][
                            cutting_result[2].index(condition)]  # 找到condition的索引，然后按该索引拿到对应的文件名
                        if condition in dcs_tested:
                            # 如果已经存在当前工况的数据，则合并已有数据和当前数据
                            file2 = dcs_individual_file[dcs_tested.index(condition)]
                            data1 = pd.read_csv(file1)
                            data2 = pd.read_csv(file2)
                            # 合并两个dataframe，自动判断按行还是按列合并
                            # 合并的两种情况：
                            # 1. 两个data的列名相同，但时间段不同，则将data2续到data1的最后一行之后，作为新行
                            # 2. 两个data的时间段相同，但列名（即参数不同，一般来自DCS系统中导出的不同的趋势组数据）不同，
                            #    则将data1和data2的列拼接起来。可能还存在更复杂的情况，目前已测试完成。
                            data2, dropped_cols = merge_two_dataframes(data1, data2)
                            # data2[data2 == 0.000] = np.nan  # dcs中的数据就有可能全是0，不能替换为np.nan，否则0不显示
                            self.dropped_dcs_cols = list(set(self.dropped_dcs_cols).union(set(dropped_cols)))
                            data2.to_csv(file2, index=False)
                            os.remove(file1)
                        else:
                            file2 = os.path.join(os.path.dirname(file1), f"{condition}{suffix_dcs_10}.csv")
                            # 否则，将当前工况的数据文件名更改为最终数据文件名，并在存在工况列表中添加当前工况
                            if os.path.exists(file2):
                                os.remove(file2)
                            os.rename(file1, file2)
                            dcs_tested.append(condition)
                            dcs_individual_file.append(file2)
                elif isinstance(cutting_result, tuple) and cutting_result[0] == "unk_slice":
                    for condition in cutting_result[2]:
                        file1 = cutting_result[1][
                            cutting_result[2].index(condition)]  # 找到condition的索引，然后按该索引拿到对应的文件名
                        if condition in unknown_tested:
                            # 如果已经存在当前工况的数据，则合并已有数据和当前数据
                            file2 = unknown_individual_file[unknown_tested.index(condition)]
                            data1 = pd.read_csv(file1)
                            data2 = pd.read_csv(file2)
                            data2, dropped_cols = merge_two_dataframes(data1, data2)
                            # data2[data2 == 0.000] = np.nan  # 有些数如循泵功率一直为0，不能替换为np.nan
                            data2.to_csv(file2, index=False)
                            os.remove(file1)
                        else:
                            file2 = os.path.join(os.path.dirname(file1), f"{condition}_剪切unk时间_修正.csv")
                            # 否则，将当前工况的数据文件名更改为最终数据文件名，并在存在工况列表中添加当前工况
                            if os.path.exists(file2):
                                os.remove(file2)
                            os.rename(file1, file2)
                            unknown_tested.append(condition)
                            unknown_individual_file.append(file2)
                elif isinstance(cutting_result, tuple) and cutting_result[0] == "kp_hollysys":
                    logger.debug("发现和利时点表文件")  # 和利时的点表文件
                    if self.point_description is None:
                        self.point_description = cutting_result[1]
                    else:
                        self.point_description.update(cutting_result[1])

        self.statusBar().showMessage('汇总试验数据...')
        self.progressBar.setValue(0)
        QApplication.processEvents()

        if len(imp_or_wsn_tested) != len(self.test_time.keys()):
            logger.warning(f"imp采集数据工况找到{len(imp_or_wsn_tested)}个，少于工况记录单记录"
                           f"工况数量{len(self.test_time.keys())}")
        if len(power_tested) != len(self.test_time.keys()):
            logger.warning(f"功率表记录工况找到{len(power_tested)}个，少于工况记录单记"
                           f"录工况数量{len(self.test_time.keys())}")
        if len(self.dropped_dcs_cols) > 0:
            logger.warning(f"dcs数据中部分列不是数值型数据，已丢弃。"
                           f"{json.dumps(self.dropped_dcs_cols, ensure_ascii=False)}｝")
        self.gather_imp_or_wsn(imp_individual_file)
        self.progressBar.setValue(33)
        QApplication.processEvents()
        self.gather_power(power_individual_file)
        self.progressBar.setValue(66)
        QApplication.processEvents()
        self.gather_dcs(dcs_individual_file)
        self.progressBar.setValue(100)
        QApplication.processEvents()
        self.gather_unknown(unknown_individual_file)
        self.statusBar().showMessage('就绪')
        self.progressBar.setVisible(False)
        self.yk_signal.emit({"title": "done", "description": ""})
        if len(self.test_time) == 0:
            logger.debug(f"未查询到工况记录单文件数据，只进行数据预处理！")
        logger.debug("数据处理完毕！")

    def button_clicked(self):
        sender = self.sender()
        if sender.text() == "保存当前项目":
            self.config.update({"水位计算点数": self.table_widget.get_value("水位计算点数"),
                                "使用表号确定压力类型": self.table_widget.get_value("使用表号确定压力类型"),
                                "未知类型文件区分文件名": self.table_widget.get_value("未知类型文件区分文件名")})
            self.project.update(self.config)
            write_as_pickle(self.project_file, self.project)
        self.statusBar().showMessage("当前项目已保存")

    def post(self, check_box: QCheckBox):
        """
        数据处理完成后的后处理阶段。

        打开计算分析软件，关闭当前界面

        :return:
        """
        from yangke.performance.data_analysis import MainWindow

        MainWindow(data_folder=self.data_folder).show()
        if check_box.isChecked():
            self.close()

    @QtCore.pyqtSlot(dict)
    def on_message_changed(self, msg):
        title = msg.get("title")
        description = msg.get("description")
        if title == "done":
            self.table_widget.get_button("自动处理").setEnabled(True)
            btn_init_cal = QPushButton("Analysis")
            q_check_box = QCheckBox("关闭当前界面")
            q_check_box.setChecked(True)
            # noinspection all
            btn_init_cal.clicked.connect(lambda: self.post(q_check_box))
            self.table_widget.setSpan(2, 3, 1, 2)
            self.table_widget.setCellWidget(2, 2, btn_init_cal)
            self.table_widget.setCellWidget(2, 3, q_check_box)

    def setting_abs_or_rel(self):
        """
        主动设置绝压/表压测点信息。设置窗口会在软件无法确认测点是绝压还是表压时自动弹出一次，之后要在此设置，需要单独调用该方法。

        :return:
        """
        self.project.update(read_from_pickle(self.project_file) or {})
        pressure_abs_specific = self.project.get("pressure_abs") or []
        pressure_rel_specific = self.project.get("pressure_rel") or []
        temp_list = pressure_abs_specific.copy()
        temp_list.extend(pressure_rel_specific)
        if len(temp_list) == 0:
            QMessageBox.warning(self,
                                "警告",
                                "待确认测点清单为空，请检查后再试！",
                                QMessageBox.Ok)
            return
            # ===================查找当前处理文件夹中是否存在额外的配置文件，配置文件中是否记录压力测点的信息======================
        # 如果没有找到测点的额外信息，则弹出窗口，提示用户确定测点是表压还是绝压
        temp_list = list(set(temp_list))
        temp_list.sort()
        dialog = ConfirmDialog(var_list=temp_list, abs_list=pressure_abs_specific, rel_list=pressure_rel_specific)

        # 让对话框以模式状态显示，即显示时QMainWindow里所有的控件都不可用，除非把dialog关闭
        dialog.setWindowModality(QtCore.Qt.ApplicationModal)
        dialog.exec()

        var_list, value_list = dialog.get_data()
        pressure_abs_specific = []  # 记录前清空两个列表
        pressure_rel_specific = []
        for var, value in zip(var_list, value_list):
            if value == "绝压":
                pressure_abs_specific.append(var)
            elif value == "表压":
                pressure_rel_specific.append(var)

        # 保存用户输入的测点信息到配置文件，便于下次启动软件是加载
        pressure_abs_specific = list(set(pressure_abs_specific))
        pressure_rel_specific = list(set(pressure_rel_specific))
        self.project.update({"pressure_abs": pressure_abs_specific, "pressure_rel": pressure_rel_specific})
        # 更新现有的绝压/表压列表
        # noinspection all
        global pressure_absolute, pressure_relative
        pressure_absolute = list(set(pressure_absolute) - set(temp_list))  # 恢复默认绝压列表
        pressure_relative = list(set(pressure_relative) - set(temp_list))  # 恢复默认表压列表
        pressure_absolute.extend(pressure_abs_specific)
        pressure_relative.extend(pressure_rel_specific)
        self.project.update(self.config)
        write_as_pickle(self.project_file, self.project)

    def ask_for_absolute_or_relative(self):
        # 查找试验测点中的压力测点是绝压还是表压
        data_series = self.temp_para

        def get_unknown_list(p_relative, p_absolute, var_list=None):
            """
            查找无法判断绝压还是表压的测点名称

            :return: 返回无法判断的测点列表
            """
            temp_list1 = []
            if isinstance(var_list, pd.Series):
                for name, _ in var_list.items():
                    if "pa" not in name.lower():  # 说明不是压力数据
                        continue
                    if "流量" in name or "差压" in name or "PD" in name:
                        continue
                    # 如果使用表号判断压力表类型
                    if self.config.get("使用表号确定压力类型"):
                        tag = name.split("_")[2]
                        if "pg" in tag.lower():
                            p_relative.append(name)
                        elif "pa" in tag.lower():
                            p_absolute.append(name)

                    if str_in_list(name, p_relative, revert=True) or str_in_list(name, p_absolute, revert=True):
                        continue
                    else:
                        temp_list1.append(name)
            else:
                for name in var_list:
                    if "pa" not in name.lower():  # 说明不是压力数据
                        continue
                    if "流量" in name or "差压" in name:
                        continue
                    # 如果使用表号判断压力表类型
                    if self.config.get("使用表号确定压力类型"):
                        tag = name.split("_")[2]
                        if "pg" in tag.lower():
                            p_relative.append(name)
                        elif "pa" in tag.lower():
                            p_absolute.append(name)

                    if str_in_list(name, p_relative, revert=True) or str_in_list(name, p_absolute, revert=True):
                        continue
                    else:
                        temp_list1.append(name)

            return temp_list1

        # noinspection all
        global pressure_absolute, pressure_relative

        temp_list = get_unknown_list(p_relative=pressure_relative, p_absolute=pressure_absolute, var_list=data_series)

        if len(temp_list) == 0:  # 如果不存在未知的测点，则直接返回
            self.temp_para = None
            return

        # ===================查找当前处理文件夹中是否存在额外的配置文件，配置文件中是否记录压力测点的信息======================
        self.project.update(read_from_pickle(self.project_file) or {})

        pressure_rel_specific = self.project.get("pressure_rel") or []
        pressure_abs_specific = self.project.get("pressure_abs") or []

        temp_list = get_unknown_list(p_relative=pressure_rel_specific, p_absolute=pressure_abs_specific,
                                     var_list=temp_list)
        pressure_absolute.extend(pressure_abs_specific)
        pressure_relative.extend(pressure_rel_specific)
        pressure_absolute = list(set(pressure_absolute))
        pressure_relative = list(set(pressure_relative))

        if len(temp_list) == 0:
            self.temp_para = None
            return
        # ===================查找当前处理文件夹中是否存在额外的配置文件，配置文件中是否记录压力测点的信息======================
        # 如果没有找到测点的额外信息，则弹出窗口，提示用户确定测点是表压还是绝压
        temp_list = list(set(temp_list))
        temp_list.sort()
        dialog = ConfirmDialog(var_list=temp_list)

        # 让对话框以模式状态显示，即显示时QMainWindow里所有的控件都不可用，除非把dialog关闭
        dialog.setWindowModality(QtCore.Qt.ApplicationModal)
        dialog.exec()

        var_list, value_list = dialog.get_data()
        for var, value in zip(var_list, value_list):
            if value == "绝压":
                pressure_abs_specific.append(var)
            elif value == "表压":
                pressure_rel_specific.append(var)

        # 保存用户输入的测点信息到配置文件，便于下次启动软件是加载
        self.project.update({"pressure_abs": pressure_abs_specific, "pressure_rel": pressure_rel_specific})
        # 更新现有的绝压/表压列表
        pressure_absolute.extend(pressure_abs_specific)
        pressure_relative.extend(pressure_rel_specific)

        # 当前线程执行完毕，将temp_para置为None，以便通知调用者当前线程结束
        self.temp_para = None


if __name__ == "__main__":
    app = QApplication(sys.argv)
    # app.setFont(QFont("Microsoft YaHei", 12))
    # app.setStyleSheet("font-size: 20px")
    w1 = MainWindow()
    w1.setFont(QFont("Microsoft YaHei", 12))
    sys.exit(app.exec_())



"""
win10系统专用，用于在windows操作系统中获取窗口、进程等应用程序操作
"""
import ctypes
import random
import string
import time

import cv2
import win32gui  # pip install pypiwin32

import win32con
import win32process
from PIL import Image
import os
import win32api
from yangke.base import pic2ndarray, crop_pic
from yangke.common.config import logger
from yangke.common.win.keyboard import get_virtual_keycode
import numpy as np

os.environ['KMP_DUPLICATE_LIB_OK'] = 'True'


def get_user32():
    """
    通过该方法返回的对象可以调用user32.dll中的函数

    :return:
    """
    # user32.dll是stdcall
    from ctypes import windll
    # import ctypes
    # _user32 = ctypes.WinDLL("user32")
    return windll.user32


def get_Ws2_32():
    """
    通过该方法返回的对象可以调用Ws2_32.dll中的函数，例如send发包函数

    :return:
    """
    from ctypes import windll
    # Ws2_32.send()  # 发包函数
    return windll.Ws2_32


def get_kernel32():
    """
    通过该方法返回的对象可以调用kernel32.dll中的函数，例如LoadLibraryA，OpenProcess等
    参见https://docs.microsoft.com/zh-cn/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocess

    :return:
    """
    # kernel32.dll是stdcall
    from ctypes import windll
    return windll.kernel32


def get_msvcrt():
    # msvcrt是微软C标准库，包含了大部分C标准函数，这些函数都是以cdecl调用协议进行调用的
    # 例如 msvcrt.printf()  msvcrt.strchr()
    from ctypes import cdll
    return cdll.msvcrt


def get_all_window():
    """
    获取windows系统中当前所有的窗口句柄及标题
    :return: 返回格式为{hwnd, title}的字典
    """
    hwnd_title = {}

    def get_all_hwnd(hwnd, mouse):
        if win32gui.IsWindow(hwnd) and win32gui.IsWindowEnabled(hwnd) and win32gui.IsWindowVisible((hwnd)):
            hwnd_title.update({hwnd: win32gui.GetWindowText(hwnd)})

    win32gui.EnumWindows(get_all_hwnd, 0)
    return hwnd_title


def find_window(title, window_cls=None, exact=True):
    """
    查找窗口句柄，可以精确查找指定标题的窗口，也可模糊查找标题包含指定字符串的窗口。
    如果精确查找，则只返回一个窗口句柄值，返回类型为int，在存在多个同名窗口时，也只返回找到的第一个。
    如果模糊查找，会返回包含指定字符串的所有窗口的句柄信息，返回格式为{hwnd:title}的字典。

    :param title: 查找的窗口名称
    :param window_cls: 窗口的类名，例如："GxWindowClass"、"Windows.UI.Core.CoreWindow"等，不指定则匹配所有类型的窗口
    :param exact: 是否精确查找，默认为True，为False是则匹配包含title的所有窗口
    :return: exact为True是返回窗口的句柄，exact为False时返回{hwnd:title}字典
    """
    # 获取句柄
    window_cls = window_cls or 0
    if exact:
        hwnd = win32gui.FindWindow(window_cls, title)
        return hwnd
    else:
        hwnd_title = get_all_window()
        hwnd_title_filter = {}
        for hwnd1, title1 in hwnd_title.items():
            title1 = str(title1)
            if title1.find(title) != -1:
                hwnd_title_filter[hwnd1] = title1
        return hwnd_title_filter


def filter_windows_by_cls(hwnd_list, cls_name=0):
    """
    按类名筛选窗口，返回类名为cls_name的窗口

    :param hwnd_list:
    :param cls_name:
    :return:
    """
    if not isinstance(cls_name, list):
        cls_name = [cls_name]
    result = []
    for hwnd in hwnd_list:
        _name = win32gui.GetClassName(hwnd)
        if _name in cls_name:
            result.append(hwnd)
    return result


def filter_windows_by_title(hwnd_list, title=None):
    if not isinstance(title, list):
        title = [title]
    result = []
    for hwnd in hwnd_list:
        _t = win32gui.GetWindowText(hwnd)
        if _t in title:
            result.append(hwnd)
    return result


def find_child_windows(hwnd_parent, window_cls=None, title=None):
    """
    获取父窗口下的子窗口

    :param hwnd_parent: 父窗口的句柄
    :param window_cls: 字符型或字符型的列表，是窗体的类名，如果需要保留多种类的窗口，则以列表形式传入多个类名
    :param title: 字符型，是窗体的类名
    :return:
    """
    if not hwnd_parent:
        return []
    window_cls = window_cls or 0
    # win32gui.FindWindowEx(hwnd_parent, hwnd_child, window_cls, title) # 该方法获取子窗口有时候获取不到，不要使用，例如不知道类名时，无法使用
    hwndChildList = []
    win32gui.EnumChildWindows(hwnd_parent, lambda hwnd, param: param.append(hwnd), hwndChildList)
    # 按类名筛选窗口
    if window_cls != 0:
        hwndChildList = filter_windows_by_cls(hwndChildList, cls_name=window_cls)
    if title is not None:
        hwndChildList = filter_windows_by_title(hwndChildList, title=title)

    return hwndChildList


def get_pid_by_hwnd(hwnd):
    """
    根据窗口句柄获取进程ID
    """
    thread_id, pid = win32process.GetWindowThreadProcessId(hwnd)
    return thread_id, pid


def get_size_of_window(hwnd, client=False):
    """
    获取窗口大小

    :param hwnd: 窗口句柄
    :param client: 是否获取客户区大小，即不包括标题框的窗口大小，默认为否，即返回总大小
    :return:
    """
    if client:
        left, top, right, bot = win32gui.GetClientRect(hwnd)
    else:
        left, top, right, bot = win32gui.GetWindowRect(hwnd)
    width = right - left
    height = bot - top
    return width, height


def capture_pic_undecorated(hwnd, x1=0, y1=0, x2=0, y2=0, save_to="memory"):
    """
    截取窗口不包括标题框的区域的画面

    :param hwnd:
    :param x1: 截图区域的左上角x坐标，
    :param y1: 截图区域的左上角y坐标
    :param x2: 截图区域的右下角x坐标
    :param y2: 截图区域的右下角y坐标
    :param save_to:
    :return:
    """
    pic = capture_pic(hwnd, save_to=save_to)
    w_, h_ = get_size_of_window(hwnd)  # 窗口总大小
    w, h = get_size_of_window(hwnd, client=True)  # 客户去大小，即不包括标题框的区域的大小
    pad_x = int((w_ - w) / 2)
    pad_y_t = int(h_ - h - pad_x)
    pad_y_b = int(h_ - h - pad_y_t)
    pic, suc = crop_pic(pic, pad_x, pad_y_t, w_ - pad_x, h_ - pad_y_b, save_to=save_to)  # 裁掉图片中窗口的修饰框
    if (x1, y1, x2, y2) != (0, 0, 0, 0):  # 如果需要传入了截图区域，则对图片进行裁剪
        pic, suc = crop_pic(pic, x1, y1, x2, y2, save_to)  # 对窗口客户区中的区域进行裁剪
    return pic


def capture_pic(hwnd, x1=0, y1=0, x2=0, y2=0, save_to: str = "memory", show=False):
    """
    截取给定句柄对应的窗口中指定位置的图像，可以处理后台画面，默认截取整个窗口画面。
    测试游戏画面可以截取成功，如逆水寒，但Windows的记事本、计算器截取不成功.

    :param hwnd: 要截图的窗口句柄
    :param x1: 截图区域的左上角x坐标，
    :param y1: 截图区域的左上角y坐标
    :param x2: 截图区域的右下角x坐标
    :param y2: 截图区域的右下角y坐标
    :param save_to: 保存截图到文件时，通过该参数指定文件名，不指定则为内存中保存的图片，否则，返回图片的ndarray数组，图片是RGB模式
    :return:
    """
    import win32ui
    left, top, right, bot = win32gui.GetWindowRect(hwnd)
    width = right - left
    height = bot - top
    # 获取句柄窗口的设备环境，覆盖整个窗口，包括非客户区，标题栏，菜单，边框
    hwnd_DC = win32gui.GetWindowDC(hwnd)
    # 创建设备描述表
    mfcDC = win32ui.CreateDCFromHandle(hwnd_DC)
    # 创建内存设备描述表
    saveDC = mfcDC.CreateCompatibleDC()
    # 创建位图对象准备保存图片
    saveBitMap = win32ui.CreateBitmap()
    # 为bitmap开辟存储空间
    saveBitMap.CreateCompatibleBitmap(mfcDC, width, height)
    # 将截图保存到saveBitMap中
    saveDC.SelectObject(saveBitMap)
    # 保存bitmap到内存设备描述表
    saveDC.BitBlt((0, 0), (width, height), mfcDC, (0, 0), win32con.SRCCOPY)

    # 保存图像
    # 方法一：windows api保存
    # saveBitMap.SaveBitmapFile(saveDC, save_to)

    # 方法二：PIL保存
    bmp_info = saveBitMap.GetInfo()
    bmp_str = saveBitMap.GetBitmapBits(True)
    im_pil = Image.frombuffer('RGB', (bmp_info['bmWidth'], bmp_info['bmHeight']), bmp_str, 'raw', 'BGRX', 0, 1)

    if save_to == "memory":  # 如果需要的是内存对象
        # noinspection all
        save_to = np.asarray(im_pil)
    else:
        im_pil.save(save_to)
    if show:
        im_pil.show()

    # 内存释放
    win32gui.DeleteObject(saveBitMap.GetHandle())
    saveDC.DeleteDC()
    mfcDC.DeleteDC()
    win32gui.ReleaseDC(hwnd, hwnd_DC)
    if (x1, y1, x2, y2) != (0, 0, 0, 0):  # 如果需要传入了截图区域，则对图片进行裁剪
        crop_pic(save_to, x1, y1, x2, y2, save_to)
    return save_to


def 注入dll(目标进程ID, dll_path):
    """
    将dll_path指定的dll文件注入到目标进程中

    :param 目标进程ID:
    :param dll_path:
    :return:
    """
    win32process.CreateRemoteThread()


def find_pic(template: str, screen: str = "temp.png", threshold: float = 0.8):
    """
    从图片中查找另一张图片的位置，找到的坐标必然为正值

    :param screen: 大图片，即在该图片中查找，如果是数组，颜色通道顺序需要为RGB
    :param template: 小图片，即在大图片中查找该图片的相对位置，如果是数组，颜色通道顺序需要为RGB
    :param threshold: 阈值，越大则要求越相似
    :return: exists, (x,y)
    """
    threshold = 1 - threshold
    try:
        # scr = cv2.imread(screen)
        scr = pic2ndarray(screen)
        tp = pic2ndarray(template)
        result = cv2.matchTemplate(scr, tp, cv2.TM_SQDIFF_NORMED)
    except cv2.error:
        print('文件错误：', screen, template)
        time.sleep(1)
        try:
            scr = cv2.imread(screen)
            tp = cv2.imread(template)
            result = cv2.matchTemplate(scr, tp, cv2.TM_SQDIFF_NORMED)
        except cv2.error:
            return False, None
    h, w = scr.shape[:2]
    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
    # logger.debug(f"{min_val=}, {max_val=}")
    # min_val越小，表示可信度越高
    exist = True
    if min_val < threshold:
        logger.debug(f"找图可信度较高:{os.path.basename(template)}")
    else:
        logger.warning(f"找图可信度较低:{os.path.basename(template)}")
        exist = False
    coordinate = (min_loc[0], min_loc[1])
    return exist, coordinate


def do_click(cx, cy, hwnd, right=False):
    """
    在窗口上点击鼠标，测试hwnd为桌面句柄时成功。
    逆水寒游戏窗口测试失败。


    :param right: 是否右键点击，默认否
    :param cx: 点击位置在窗口的相对x坐标
    :param cy: 点击位置在窗口的相对y坐标
    :param hwnd: 需要点击的窗口句柄
    :return:
    """
    logger.debug(f"左键点击({cx, cy})")
    long_position = win32api.MAKELONG(cx, cy)
    if right:
        win32api.PostMessage(hwnd, win32con.WM_RBUTTONDOWN, win32con.MK_RBUTTON, long_position)
        time.sleep(0.1)
        win32api.PostMessage(hwnd, win32con.WM_RBUTTONUP, win32con.MK_RBUTTON, long_position)
    else:
        # win32api.PostMessage(hwnd, win32con.WM_LBUTTONDOWN, win32con.MK_LBUTTON, long_position)
        # time.sleep(0.1)
        # win32api.PostMessage(hwnd, win32con.WM_LBUTTONUP, win32con.MK_LBUTTON, long_position)
        win32api.SendMessage(hwnd, win32con.WM_LBUTTONDOWN, win32con.MK_LBUTTON, long_position)
        time.sleep(0.1)
        win32api.SendMessage(hwnd, win32con.WM_LBUTTONUP, win32con.MK_LBUTTON, long_position)


def key_press(hwnd, key):
    ...


def post_key(hwnd, key, last_time=0.01):
    """
    向指定窗口发送按键事件。
    记事本已测试成功，需要注意的是hwnd需要只想记事本的编辑区域的子窗口。
    逆水寒测试成功。但部分按键的功能与实际按键功能不一致，原因未知。
    天谕测试部分成功，单个按键功能正常，但是组合按键中的shift、Alt和Ctrl失效，原因未知。
    某些应用可能无法使用，测试失败的有：斗鱼客户端(客户端会激活，但功能无效)

    示例：
    hwnd = 记事本编辑框的窗口句柄 # 需要注意的是，如果是Ctrl+S这类功能按键，窗口句柄是记事本的根句柄还是下面的字句并都无所谓，但如果是编辑内容的输入，如输入A字母，则必须是编辑子窗口的句柄
    post_key(hwnd, 'Ctrl+A')  # 全选记事本内容，相当于发送Ctrl+A按键事件
    post_key(hwnd, 'Ctrl+S')  # 打开记事本设置面板，相当于发送Alt+S按键事件
    post_key(hwnd, 'A')  # 在编辑区输入字母a，相当于在编辑区按“A"键
    post_key(hwnd, 'escape')  # 按Esc键

    虚拟键码参加以下链接
    https://learn.microsoft.com/zh-cn/windows/win32/inputdev/virtual-key-codes

    :param hwnd: 窗口句柄。
    :param key: 需要点击的普通键，如'A', '1'等，具体参见keyboard.py中的VkCode的键值
    :param last_time: 按键持续时间，即键位按下的时间，单位为s
    :return:
    """
    """
    微软PostMessage函数原型为：
    BOOL PostMessageW(HWND hwnd, UINT Msg, WPARAM wParam, LPARAM lParam)
    其中hWnd是窗口句柄，Msg是消息的ID，wParam和lParam都需要按照消息的要求进行设置。每个消息的要求都不一样，具体如何设置，需要查看对应消息的文档。
    """
    _user32 = get_user32()
    PBYTE256 = ctypes.c_ubyte * 256
    time_last = last_time + random.randint(1, 10) / 100

    special_key = False  # 不知道有什么用，Ctrl、Alt、Shift等按键都不需要设置该参数

    # -------------------------- 将按键字符串转换为PostMessage可以识别的参数 -----------------------------------
    addition_key = None  # 附加的功能按键
    if '+' in key:
        keys = key.split('+')
        for k in keys:
            if k.lower() == "ctrl":
                if addition_key is None or addition_key == []:
                    addition_key = [win32con.VK_CONTROL]
                else:
                    addition_key.append(win32con.VK_CONTROL)
            elif k.lower() == 'shift':
                if addition_key is None or addition_key == []:
                    addition_key = [win32con.VK_SHIFT]
                else:
                    addition_key.append(win32con.VK_SHIFT)
            elif k.lower() == 'alt':
                if addition_key is None or addition_key == []:
                    addition_key = [win32con.VK_MENU]
                else:
                    addition_key.append(win32con.VK_MENU)
            else:
                key = k
    key = get_virtual_keycode(key)  # 虚拟键码等于字符的ASCII码值
    # -------------------------- 将按键字符串转换为PostMessage可以识别的参数 -----------------------------------

    if win32gui.IsWindow(hwnd):
        thread_id = _user32.GetWindowThreadProcessId(hwnd, None)  # win32process.GetWindowThreadProcessId给出了错误的值

        lparam = win32api.MAKELONG(0, _user32.MapVirtualKeyA(key, 0))  # 将虚拟按键code转换成长整形
        # lparam_ctrl = win32api.MAKELONG(0, MapVirtualKeyW(win32con.VK_CONTROL, 0)) | 0x00000001
        msg_down = win32con.WM_KEYDOWN
        msg_up = win32con.WM_KEYUP

        if special_key:
            lparam = lparam | 0x1000000

        # 如果有修改-使用PostMessage和AttachThreadInput
        pKeyBuffers = PBYTE256()
        pKeyBuffers_old = PBYTE256()

        win32gui.SendMessage(hwnd, win32con.WM_ACTIVATE, win32con.WA_ACTIVE, 0)  # 激活hwnd窗口

        # ------------------------- 如果是组合按键，才需要设置键盘状态 --------------------------------
        if addition_key is not None and len(addition_key) > 0:
            # 发送组合键时需要AttachThreadInput方法
            _user32.AttachThreadInput(win32api.GetCurrentThreadId(), thread_id, True)  # 附加python线程至目标窗口
            # 记录当前按下的组合键，如ctrl、shift、alt的状态，状态记录在pKeyBuffers_old中
            _user32.GetKeyboardState(ctypes.byref(pKeyBuffers_old))

            # 记录需要按下的组合键，如ctrl、shift、alt的状态，状态记录在pKeyBuffers中
            for mod_key in addition_key:
                if mod_key == win32con.VK_MENU:  # Alt键需要单独处理，VK_MENU就是ALT键
                    lparam = lparam | 0x20000000
                    msg_down = win32con.WM_SYSKEYDOWN
                    msg_up = win32con.WM_SYSKEYUP
                pKeyBuffers[mod_key] |= 128
            _user32.SetKeyboardState(ctypes.byref(pKeyBuffers))  # 将当前的键盘状态设置为记录的pKeyBuffers
            time.sleep(time_last)
        # win32api.PostThreadMessage(ThreadId, msg_down, key, lparam)
        win32api.PostMessage(hwnd, msg_down, key, lparam)
        time.sleep(time_last)
        # win32api.PostThreadMessage(ThreadId, msg_up, key, lparam | 0xC0000000)
        win32api.PostMessage(hwnd, msg_up, key, lparam | 0xC0000000)
        time.sleep(time_last)

        # win32api.PostMessage(hwnd, win32con.WM_SYSKEYDOWN, win32con.VK_MENU, 0x20380001)
        # win32api.PostMessage(hwnd, win32con.WM_SYSKEYDOWN, 0x56, 0x20200001)  # 0x56是V键
        # win32api.PostMessage(hwnd, win32con.WM_SYSCHAR, 0x76, 0x20200001)
        # win32api.PostMessage(hwnd, win32con.WM_SYSKEYUP, 0x56, 0xE0200001)
        # win32api.PostMessage(hwnd, win32con.WM_KEYUP, win32con.VK_MENU, 0xC0380001)
        # win32api.PostMessage(hwnd, win32con.WM_SYSKEYDOWN, 0x56, 1 << 29)  # 第29位为1表示alt键按下

        # ------------------------- 如果是组合按键，恢复设置键盘状态 --------------------------------
        if addition_key is not None and len(addition_key) > 0:
            _user32.SetKeyboardState(ctypes.byref(pKeyBuffers_old))
            time.sleep(0.02)
            _user32.AttachThreadInput(win32api.GetCurrentThreadId(), thread_id, False)  # 从目标线程上分离python线程


def send_message(hwnd, str):
    """
    想窗口发送字符串，窗口必须是可以接受字符串的，例如记事本的编辑区。测试成功。
    需要注意的是，该方法不是发送按键事件的方法。

    :param hwnd:
    :param str:
    :return:
    """
    str2int = [ord(c) for c in str]
    for x in str2int:
        win32api.SendMessage(hwnd, win32con.WM_CHAR, x, 0)


def get_nsh_window():
    """
    获取逆水寒的游戏窗口
    :return:
    """
    hwnd_title = find_window('逆水寒 角色ID', exact=False)
    if len(hwnd_title) > 0:
        hwnd, title = hwnd_title.popitem()
        return hwnd
    else:
        return None


if __name__ == "__main__":
    # hwnd = get_nsh_window()
    # capture_pic(hwnd, show=False, save_to="temp.png")
    # post_key(hwnd, ord('F'), [win32con.VK_CONTROL])
    from yangke.common.win.keyboard import *

    hwnd = find_window("记事本", exact=False)
    hwnd = list(hwnd.keys())[0]
    post_key(hwnd, ord("A"), [])
    capture_pic(hwnd)
    while True:
        click_key(hwnd, "V")
# [722220, 394318]



import datetime
import os.path
from enum import unique, Enum
from shutil import copyfile

from yangke.base import get_key_value, yield_every_day
from yangke.common.fileOperate import fill_excel_cell
from yangke.sis.export_history_values import get_history


class Report:
    def __init__(self, template, tags, report_type="year", cache_folder="cache"):
        """
        报表类，用于生成基于Excel的报表

        tags示例：
# ----------------------------------------------------------
from enum import unique, Enum
from yangke.base import get_key_value
@unique
@get_key_value
class TagsRead(Enum):
    发电机功率1 = "DCS1:LOAD"
    总燃料量1 = "DCS1:TFFREAL"
tags = TagsRead
# ----------------------------------------------------------

        :param template: 报表的模板文件，一般为excel文件
        :param tags: 报表用到的数在RDBProxy代理服务器上的测点信息
        :param report_type: 报表类型，可取值'year', 'month'，表示报表需要的数据的范围，year表示需要一年的数据，month表示需要一月的数据
        :param cache_folder: 缓存文件夹
        """
        self.template = template
        self.out_file = None  # 报表的输出文件，由self.generate_report_excel()方法指定
        self.tags = tags
        self.report_type = report_type
        self.cache_folder = os.path.abspath(cache_folder)
        self.year = None
        self.month = None
        self.week = None
        self.day = None
        self.rules = []
        if not os.path.exists(self.cache_folder):
            os.makedirs(self.cache_folder)

    def generate_report_excel(self, out_file, year=None, month=None, week=None, day=None):
        """
        生成报表文件

        :param out_file: 报表保存的文件
        :param year: 年份，如果不指定，则默认生成今年
        :param month: 如果是月报，则需要指定月份，如果不指定，则默认当月
        :param week:
        :param day:
        :return:
        """

        now = datetime.datetime.now()
        if self.report_type == "year":
            self.year = year or now.year
        elif self.report_type == "month":
            self.year = year or now.year
            self.month = month or now.month
        elif self.report_type == "week":
            self.year = year or now.year
            _week = int(now.strftime('%W'))  # %W取一年中的星期数（00-53），星期一为星期的开始，%w星期天为星期的开始
            self.week = week or _week
        elif self.report_type == "day":
            self.year = year or now.year
            self.month = month or now.month
            self.day = day or now.day  # 月中的天数{1-31}，不是年中的天数

        start_time, end_time = None, None
        if self.report_type == "year":  # 如果是年报，应该只有year参数是有数的
            start_time = f"{self.year}-1-1 00:00:00"
            end_time = now if self.year >= now.year else f"{self.year}-12-31 23:59:59"
        elif self.report_type == "month":  # 如果是月报
            start_time = f"{self.year}-{self.month}-1 00:00:00"
            if self.year == now.year and self.month == now.month:
                end_time = now
            else:
                end_time = f"{self.year}-{self.month}- 23:59:59"  # 某月的最后一天

        self.download_data(start_time, end_time)

        # 复制模板文件作为报表文件的输出文件
        copyfile(self.template, out_file)
        self.out_file = out_file
        self.exec_rules()  # 执行报表生成规则，填充报表

        return self.out_file

    def add_fill_rules(self, excel_cell=None, content: str | dict = "test hello world"):
        """
        excel文件的填充规则。

        content内容可以是参数组成的公式，参数必须在self.tags中存在。示例如下：
        content = {"expression": "累积值(发电机功率1, 2022-1-1)"}表示填充内容是2022年1月1日发电机功率1的累积值。因为
        数据库中功率单位一般为MW，则累积值单位是MWs，换算成电量kWh应该除以3.6，所以，如果需要填入的是当天的电量，则：
        content = content={"expression": "累积值(燃料消耗量)/3600", "date_range": "{year}年1月"}

        :param excel_cell: {"sheet": 0, "cell": "A1"}表示第一个sheet的A1单元格
        :param content: 需要填充的内容
        :return:
        """
        if excel_cell is None:
            excel_cell = {"sheet": 0, "cell": "A1"}
        if content is None:
            return
        self.rules.append([excel_cell, content])

    def exec_rules(self):
        for cell, content in self.rules:
            # 计算填充内容
            exp = content.get("expression")

            _ = fill_excel_cell(self.out_file, sheet=cell.get("sheet"), cell=cell.get("cell"), value=val, visible=True,
                                close=False)
            _ = fill_excel_cell(self.out_file, )

    def download_data(self, start_time, end_time):
        """
        下载指定时间段的数据，默认的时间间隔是1分钟，会检测本地缓存数据是否存在，只下载没有缓存的数据

        :param start_time:
        :param end_time:
        :return:
        """
        if isinstance(start_time, str):
            start_time = datetime.datetime.strptime(start_time, '%Y-%m-%d %H:%M:%S')
        if isinstance(end_time, str):
            end_time = datetime.datetime.strptime(end_time, '%Y-%m-%d %H:%M:%S')
        # 判断需要下载时间段的数据是否存在
        non_exist_days = []  # 没有缓存数据的日期的列表，每一个元素对应一天
        if self.report_type == "year":
            # 如果是年报，则检查本年度每天的数据是否存在
            year = start_time.year
            for day in yield_every_day(start_time.year):
                if end_time < day:  # 当天的数据也会统计
                    break
                data_file = os.path.join(self.cache_folder, f"{year}_{day.month}_{day.day}.csv")
                if not os.path.exists(data_file):
                    non_exist_days.append(day)
        elif self.report_type == "month":
            # 如果是月报，则检查本月每天的数据是否存在
            year = start_time.year
            month = start_time.month
            for day in yield_every_day(year, month):
                if end_time < day:
                    break
                data_file = os.path.join(self.cache_folder, f"{year}_{day.month}_{day.day}.csv")
                if not os.path.exists(data_file):
                    non_exist_days.append(day)

        for day in non_exist_days:  # 遍历每一天，逐步下载每一天的数据
            _year, _month, _day = day.year, day.month, day.day
            _s = datetime.datetime(year=_year, month=_month, day=_day, hour=0, minute=0, second=0)
            _e = datetime.datetime(year=_year, month=_month, day=_day, hour=23, minute=59, second=59)
            df = get_history(self.tags.get_values(), self.tags.get_keys(), start_time=_s, end_time=_e)
            if df.shape[0]:  # 如果最后一条数据的时间等于23：59分则说明当天的数据是完整的，否则不完整，不予保存
                df.to_csv(os.path.join(self.cache_folder, f"{_year}_{_month}_{_day}.csv"))

    def clear_cache(self, day=None):
        """
        清理缓存数据，已下载的数据会缓存在本地，如果本地数据和服务器不一致，可以通过该方法清理缓存后再次下载，否则，self.download_data()方法只是
        加载本地的缓存数据，而不是真正的下载服务器数据。

        :return:
        """
        if day is None:
            os.removedirs(self.cache_folder)
        else:
            os.remove(os.path.join(self.cache_folder, f"{day.year}_{day.month}_{day.day}.csv"))  # 只删除某一天的文件



"""
lib4python库的总入口，所有的子模块可以由此处启动
"""
import sys
import os

from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QAction, QMainWindow, QLabel

from yangke.base import get_settings
from yangke.common.qt import YkWindow, set_menu_bar
from yangke.common.config import logger

# 国际化
import gettext
from babel import Locale

gettext.install('main', 'locale')  # 将 _() 函数全局安装到内建命名空间中，使得所有模块都可以调用
locale = Locale.parse('zh_CN')


def init_get_text(domain=None):
    """
    国际化设置，使得各国语言包可以加载

    :param domain:
    :return:
    """
    if domain is None:
        domain = os.path.basename(__file__).split(".")[0]
    locale_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'locale')
    gettext.bindtextdomain(domain, locale_dir)
    gettext.textdomain(domain)
    gettext.find(domain, "locale", languages=["zh_CN", "en_US"])
    return gettext.gettext


class MainFrame(YkWindow):
    def __init__(self, setting_file=os.path.abspath("settings.yaml")):
        # super(MainFrame, self).__init__(setting_file)
        super(MainFrame, self).__init__()
        self.w = None

    def init_ui(self):
        set_menu_bar(self, "config/ui_menu.yaml")

    def init_mysql(self):
        super(MainFrame, self).init_mysql(ui_file=os.path.abspath("config/ui_mysql_config.yaml"))

    def start_program(self):
        """
        根据菜单项启动对应的软件

        :return:
        """
        if isinstance(self.sender(), QAction):
            if 'HPD' in self.sender().text():
                command = "start_hpd"
            elif self.sender().text() == "DP(性能试验)":
                command = "start_dp"
            elif self.sender().text() == "热力系统图":
                command = "start_system_figure"

        if command == "start_dp":
            from yangke.performance.data_process import MainWindow
            self.w = MainWindow()
        elif command == "start_hpd":
            from yangke.performance.heat import MainWindow
            self.w = MainWindow()
        elif command == "start_system_figure":
            self.w = QMainWindow(parent=self)  # 共享父类的app.exec_方法
            self.w.setCentralWidget(QLabel("测试"))
            self.w.show()

        self.w.setFont(QFont("Microsoft YaHei", 12))


init_get_text()
if __name__ == "__main__":
    app = QApplication(sys.argv)
    w1 = MainFrame()
    sys.exit(app.exec_())



"""
寻路算法
"""
import pygame
import math
from enum import Enum
import random


def point_index_rectangles(x, y, grid: dict):
    """
    判断点所在的区域的行索引和列索引
    :param grid:
    :return: row_index, col_index
    """
    for pos, rect in grid.items():
        if x == rect[0] or x == rect[0] + rect[2]:
            return x, -1  # 说明点在平行于y轴的横线上
        if y == rect[1] or y == rect[1] + rect[3]:
            return -1, y  # 说明点在平行于x轴的横线上
        if pygame.rect.Rect(rect).collidepoint(x, y):
            return pos  # 点所在的矩形的索引

    return -1, -1  # 说明点不在任何grids中的区域中


def search_until_obstacle(pos=(0, 0), direction="up", obstacle_grid=None, available_grid=None):
    """
    从某个位置向上搜索，直到遇到障碍物，返回起始位置和结束位置的地图区域索引

    :param pos:
    :param direction:
    :param obstacle_grid:
    :param available_grid:
    :return: row1, col1, row2, col2
    """
    start_x, start_y = pos
    row, col = point_index_rectangles(pos[0], pos[1], obstacle_grid)
    available_row = 0
    if direction == "up":
        if row == -1 or col == -1:  # 说明起点不在障碍物区域外部或边界上
            row, col = point_index_rectangles(start_x, start_y - 1, available_grid)
            row1 = row
            col1 = col
            while row != -1 or col != -1:
                # 不是障碍物
                temp_up_bound = available_grid.get((row, col))
                if temp_up_bound is not None:
                    temp_up_bound = temp_up_bound[1]
                if row != -1:
                    available_row = row
                row, col = point_index_rectangles(start_x, temp_up_bound - 1, available_grid)
            return row1, col1, available_row, col
        else:
            print("搜索起始点位于障碍物区域内")
            return -1, -1, -1, -1


def generate_a_star_map(available_grid, obstacle_grid):
    """
    根据可通行区域和障碍物区域的索引及绝对坐标，计算a_star_map的大小和障碍物列表

    :param available_grid:  索引+坐标，例如{(0, 0): [0, 0, 59, 190], (1, 0): [0, 190, 59, 49], ...}
    :param obstacle_grid: 障碍物的索引+坐标，行列的索引+索引对应区域的矩形位置及大小。
    :return: 地图的大小和障碍物区域
    """
    all_keys = []
    all_keys.extend(list(available_grid.keys()))
    all_keys.extend(list(obstacle_grid.keys()))
    row, col = -1, -1  # 地图划分的行数和列数
    for k in all_keys:
        x, y = k
        row = max(x, row)
        col = max(y, col)
    return (row, col), list(obstacle_grid.values())


def coor_2_index(start_x, start_y, grid) -> tuple:
    """
    根据点在画布上的绝对坐标判断其属于画布上的那个区域，返回区域的索引

    :param start_x:
    :param start_y:
    :param grid:
    :return:
    """
    for k, v in grid.items():
        left, right, top, bottom = v[0], v[0] + v[2], v[1], v[1] + v[3]
        if left <= start_x <= right and top <= start_y <= bottom:
            return k
    return -1, -1


def find_connect_line(start_position, start_direction, end_position, end_direction, obstacles=None, bound=None,
                      available_grid=None, obstacle_grid=None):
    """
    查找最优的连接路径

    :param obstacle_grid:
    :param available_grid:
    :param start_position: 路径起点坐标
    :param start_direction: 路径起点的方向，即起点朝哪个方向是正向，一般部件上的锚点有上下左右四个方向，假如是上方锚点，则连接线朝上时少一个拐点，连接线朝其他方向时多一个锚点
    :param end_position: 路径终点坐标
    :param end_direction: 路径终点的方向
    :param obstacles: 障碍物的rectangle列表
    :param bound: 边界
    :return:
    """
    result = []
    if available_grid is None or obstacle_grid is None:
        available_grid, obstacle_grid = generate_obstacle_map(bound, obstacles)

    if start_position[0] > end_position[0]:  # 如果起点在右侧，则交换起点和终点位置
        start_position, end_position = end_position, start_position
        start_direction, end_direction = end_direction, start_direction
    start_x, start_y = start_position
    end_x, end_y = end_position

    map_size, block_list = generate_a_star_map(available_grid, obstacle_grid)
    pos_s_node = coor_2_index(start_x, start_y, obstacle_grid)  # 点必然在障碍物区域上，因此第三个参数无需传入可通行区域的信息
    pos_e_node = coor_2_index(end_x, end_y, obstacle_grid)
    a_map = Map(map_size)  # Map只需要地图大小即可
    my_a_star = AStar(map_size, pos_s_node, pos_e_node)
    my_a_star.set_block(block_list)
    if my_a_star.run() == 0:
        route_list = my_a_star.get_route()
        return route_list
    else:
        print('路径规划失败！')


def generate_obstacle_map(bound: list, obstacles: list):
    """
    生成障碍物地图，必须保证bound包含所有的obstacles，即obstacles不能超出边界的范围

    :param bound: 地图边界，[x,y,w,h]
    :param obstacles: 障碍物列表，每一个障碍物是一个pygame.Rect对象
    :return: 返回两个字典，第一个为路径可通过的格子，第二个为障碍物的格子，字典item项格式为 {(格子行号, 格子列号): [x,y,w,h]}
    """
    bound_x, bound_y, bound_w, bound_h = bound
    rectangles = {}  # {[x,y,w,h]}
    x_list = []
    y_list = []
    for ob in obstacles:
        x, y, w, h = ob
        x_list.extend([x, x + w])
        y_list.extend([y, y + h])
    x_list = list(set(x_list))  # 去重
    x_list.sort()
    y_list = list(set(y_list))
    y_list.sort()
    x_list.append(bound_x + bound_w)
    y_list.append(bound_y + bound_h)
    temp_x_start = bound_x
    for i, x in enumerate(x_list):
        temp_y_start = bound_y
        for j, y in enumerate(y_list):
            rectangles.update({(j, i): [temp_x_start, temp_y_start, x - temp_x_start, y - temp_y_start]})
            temp_y_start = y
        temp_x_start = x

    # 从rectangles中标记出障碍物所在的格子
    available_rectangles = {}
    obstacles_rectangles = {}
    for pos, rect in rectangles.items():
        x, y, w, h = rect
        if pygame.rect.Rect(x, y, w, h).collidelist(obstacles) != -1:  # collidelist返回碰撞的索引
            obstacles_rectangles.update({pos: rect})
        else:
            available_rectangles.update({pos: rect})

    return available_rectangles, obstacles_rectangles


class Map(object):
    def __init__(self, map_size):
        self.map_size = map_size

    def generate_cell(self, cell_width, cell_height):
        x_cell = -cell_width
        for num_x in range(self.map_size[0] // cell_width):
            y_cell = -cell_height
            x_cell += cell_width
            for num_y in range(self.map_size[1] // cell_height):
                y_cell += cell_height
                yield (x_cell, y_cell)


class Node:
    def __init__(self, pos):
        self.pos = pos
        self.father = None
        self.g = 0
        self.f = 0

    def compute_fx(self, encode, father):
        """
        节点的目标函数值f，其值为 f=g+h。g(x)是指起点到当前节点的实际最小距离，h(x)是指从当前节点到目标节点的估计最小值

        :param encode:
        :param father:
        :return:
        """
        if father is None:
            print("未设置当前节点的父节点")

        gx_father = father.g
        # 采用欧式距离计算当前节点与父节点的距离
        # gx_f2n = math.sqrt((father.pos[0] - self.pos[0]) ** 2 + (father.pos[1] - self.pos[1]) ** 2)
        gx_f2n = abs(father.pos[0] - self.pos[0]) + abs(father.pos[1] - self.pos[1])
        g = gx_f2n + gx_father  # 当前节点到起点的距离=父节点到起点的距离+当前节点到父节点的距离

        # hx_n2encode = math.sqrt((self.pos[0] - encode.pos[0]) ** 2 + (self.pos[1] - encode.pos[1]) ** 2)
        hx_n2encode = abs(self.pos[0] - encode.pos[0]) + abs(self.pos[1] - encode.pos[1])
        f = g + hx_n2encode
        return g, f

    def set_fx(self, end_code, father):
        self.g, self.f = self.compute_fx(end_code, father)
        self.father = father

    def update_fx(self, encode, father):
        g, f = self.compute_fx(encode, father)
        if f < self.f:
            self.g, self.f = g, f
            self.father = father


class AStar:
    def __init__(self, map_size, pos_sn, pos_en):
        self.map_size = map_size
        self.open_list = []  # open_list中储存待考察的节点
        self.close_list = []  # 已经考察过的节点
        self.block_list = []
        self.s_node = Node(pos_sn)
        self.e_node = Node(pos_en)
        self.c_node = self.s_node  # 当前搜索到的节点

    def run(self):
        self.open_list.append(self.s_node)
        while len(self.open_list) > 0:
            # 查找open_list中f最小的点
            self.open_list.sort(key=lambda x: x.f)
            self.c_node = self.open_list[0]
            del self.open_list[0]
            self.close_list.append(self.c_node)

            # 扩展当前f最小的节点，并进入下一次循环搜索
            self.extend(self.c_node)
            # 如果open_list列表为空，或者当前搜索节点为目标节点，则跳出循环
            if len(self.open_list) == 0 or self.c_node.pos == self.e_node.pos:
                break

        if self.c_node.pos == self.e_node.pos:
            self.e_node.father = self.c_node.father
            return 0
        else:
            return -1

    def extend(self, c_node):
        """
        计算当前节点的相邻节点的距离，如果其相邻接点已经计算过了，则更新相邻节点的距离信息

        :param c_node:
        :return:
        """
        nodes_neighbor = self.get_neighbor(c_node)
        for node in nodes_neighbor:
            # 判断节点node是否在close_list和block_list中
            if node.pos in list(map(lambda x: x.pos, self.close_list)) or node.pos in self.block_list:
                continue
            else:
                if node.pos in list(map(lambda x: x.pos, self.open_list)):  # 如果节点在open中，则更新该节点的距离信息
                    node.update_fx(self.e_node, c_node)
                else:
                    node.set_fx(self.e_node, c_node)  # 否则，计算节点的距离，并将节点加入open_list中
                    self.open_list.append(node)

    def set_block(self, block_list):
        self.block_list.extend(block_list)

    def get_neighbor(self, node):
        """
        拿到节点node的相邻节点，会剔除超出地图边界的节点索引

        :param node:
        :return:
        """
        offsets = [(-1, 0), (1, 0), (0, -1), (0, 1)]
        nodes_neighbor = []
        x, y = node.pos[0], node.pos[1]
        for os in offsets:
            x_new, y_new = x + os[0], y + os[1]
            pos_new = (x_new, y_new)
            # 判断是否在地图范围内，超出范围则跳过
            if x_new < 0 or x_new > self.map_size[0] - 1 or y_new < 0 or y_new > self.map_size[1] - 1:
                continue
            nodes_neighbor.append(Node(pos_new))
        return nodes_neighbor

    def get_route(self):
        route = []
        current_node = self.e_node

        while True:
            route.append(current_node.pos)
            current_node = current_node.father
            if current_node.pos == self.s_node.pos:
                break

        route.append(self.s_node.pos)
        route.reverse()
        return route


CELL_WIDTH = 16  # 单元格宽度
CELL_HEIGHT = 16  # 单元格长度
BORDER_WIDTH = 1  # 边框宽度
BLOCK_NUM = 50  # 地图中的障碍物数量
from random import randint


class Color(Enum):
    ''' 颜色 '''
    RED = (255, 0, 0)
    GREEN = (0, 255, 0)
    BLUE = (0, 0, 255)
    WHITE = (255, 255, 255)
    BLACK = (0, 0, 0)

    @staticmethod
    def random_color():
        '''设置随机颜色'''
        r = randint(0, 255)
        g = randint(0, 255)
        b = randint(0, 255)
        return (r, g, b)


def main(map_size=(80, 40), start_node=(1, 1), end_node=(30, 20), block_list=None):
    if block_list is None:
        block_list = [(3, 1)]
    map_size = map_size
    pos_s_node = start_node
    pos_e_node = end_node
    my_a_star = AStar(map_size, pos_s_node, pos_e_node)
    my_a_star.set_block(block_list)
    if my_a_star.run() == 0:
        route_list = my_a_star.get_route()
        print(route_list)
        show_result(map_size, pos_s_node, pos_e_node, block_list, route_list)
    else:
        print('路径规划失败！')


def gen_blocks(width, height):
    """
    随机生成障碍物
    :param width: 地图宽度
    :param height: 地图高度
    :return:返回障碍物坐标集合
    """
    i, block_list = 0, []
    random.seed(100)
    while i < BLOCK_NUM:
        block = (randint(0, width - 1), randint(0, height - 1))
        if block not in block_list:
            block_list.append(block)
            i += 1

    return block_list


def show_result(map_size, pos_sn, pos_en, block_list, route_list):
    # 初始化导入的Pygame模块
    pygame.init()
    # 此处要将地图投影大小转换为像素大小，此处设地图中每个单元格的大小为CELL_WIDTH*CELL_HEIGHT像素
    mymap = Map((map_size[0] * CELL_WIDTH, map_size[1] * CELL_HEIGHT))
    pix_sn = (pos_sn[0] * CELL_WIDTH, pos_sn[1] * CELL_HEIGHT)
    pix_en = (pos_en[0] * CELL_WIDTH, pos_en[1] * CELL_HEIGHT)
    # 对block_list和route_list中的坐标同样要转换为像素值
    bl_pix = list(map(transform, block_list))
    rl_pix = list(map(transform, route_list))
    # 初始化显示的窗口并设置尺寸
    screen = pygame.display.set_mode(mymap.map_size)
    # 设置窗口标题
    pygame.display.set_caption('A*算法路径搜索演示：')
    # 用白色填充屏幕
    screen.fill(Color.WHITE.value)  # 为什么用参数Color.WHITE不行？

    # 绘制屏幕中的所有单元格
    for (x, y) in mymap.generate_cell(CELL_WIDTH, CELL_HEIGHT):
        if (x, y) in bl_pix:
            # 绘制黑色的障碍物单元格，并留出2个像素的边框
            pygame.draw.rect(screen, Color.BLACK.value, (
                (x + BORDER_WIDTH, y + BORDER_WIDTH), (CELL_WIDTH - 2 * BORDER_WIDTH, CELL_HEIGHT - 2 * BORDER_WIDTH)))
        else:
            # 绘制绿色的可通行单元格，并留出2个像素的边框
            pygame.draw.rect(screen, Color.GREEN.value, (
                (x + BORDER_WIDTH, y + BORDER_WIDTH), (CELL_WIDTH - 2 * BORDER_WIDTH, CELL_HEIGHT - 2 * BORDER_WIDTH)))
    # 绘制起点和终点
    pygame.draw.circle(screen, Color.BLUE.value, (pix_sn[0] + CELL_WIDTH // 2, pix_sn[1] + CELL_HEIGHT // 2),
                       CELL_WIDTH // 2 - 1)
    pygame.draw.circle(screen, Color.RED.value, (pix_en[0] + CELL_WIDTH // 2, pix_en[1] + CELL_HEIGHT // 2),
                       CELL_WIDTH // 2 - 1)

    # 绘制搜索得到的最优路径
    rl_pix = [(pos[0] + CELL_WIDTH // 2, pos[1] + CELL_HEIGHT // 2) for pos in rl_pix]
    pygame.draw.aalines(screen, Color.RED.value, False, rl_pix)
    keep_going = True
    while keep_going:
        pygame.time.delay(100)
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                keep_going = False
        pygame.display.flip()


def transform(pos):
    x_new, y_new = pos[0] * CELL_WIDTH, pos[1] * CELL_HEIGHT
    return x_new, y_new


if __name__ == '__main__':
    main()



# dm_ret = dm.BindWindowEx(hwnd,"dx.graphic.opengl","dx.mouse.position.lock.api|dx.mouse.position.lock.message|dx.mouse.clip.lock.api|dx.mouse.input.lock.api|dx.mouse.state.api|dx.mouse.api|dx.mouse.cursor","windows","dx.public.active.api|dx.public.active.message|dx.public.hide.dll|dx.public.active.api2|dx.public.anti.api|dx.public.km.protect|dx.public.inject.super|dx.public.memory|dx.public.inject.c",11)
# 本模块只能使用32位python调用
import time

import requests

from yangke.common.config import logger
from yangke.core import runCMD
import win32com.client
import ctypes
from xmlrpc.server import SimpleXMLRPCServer
from socketserver import ThreadingMixIn

dm = None


# http://localhost:8888/SetKey/mmdlljafd8c5cbf193e99306e9eb61ceb5bd44/ewdSEKFZP

# http://localhost:8888/Bind/
class DMRemote:
    def __init__(self, key="mmdlljafd8c5cbf193e99306e9eb61ceb5bd44", add_key="ewdSEKFZP", port=8888):
        if add_key is None:
            add_key = ""
        try:
            # res = runCMD(
            #     f'C:\\Users\\54067\\PycharmProjects\\RestDmMouseKey\\dist\\RestDM.exe "{key}" "{add_key}" {port}',
            #     wait_for_result=False)
            # # logger.debug(res)
            logger.debug(f"需要手动启动RestDM.exe软件")
            time.sleep(1)
        except:
            logger.warning(f"自动启动RestDM.exe失败，请确保该应用正常运行！")
        self.port = port

    def DmGuard(self, enable, type_="memory4"):
        res = requests.get(url=f"http://localhost:{self.port}/DmGuard/{type_}")
        print(res.text)

    def bind_window(self, hwnd, display="gdi2",
                    mouse="dx.mouse.position.lock.api|dx.mouse.position.lock.message|dx.mouse.clip.lock.api|dx.mouse.input.lock.api|dx.mouse.state.api|dx.mouse.api|dx.mouse.cursor",
                    keypad="windows",
                    public="",
                    mode=11):
        """
        设置大漠的窗口绑定方式，参考BindWindow参数说明
        dm_ret = dm.BindWindowEx(hwnd,"gdi2","dx.mouse.position.lock.api|dx.mouse.position.lock.message|dx.mouse.clip.lock.api|dx.mouse.input.lock.api|dx.mouse.state.api|dx.mouse.api|dx.mouse.cursor","windows","",11)

        """
        params = {
            "hwnd": hwnd,
            "display": display,
            "mouse": mouse,
            "keypad": keypad,
            "public": public,
            "mode": mode
        }
        res = requests.post(url=f"http://localhost:{self.port}/BindWindow/", params=params)
        print(res.text)

    def unbind_window(self):
        _ = requests.post(url=f"http://localhost:{self.port}/UnBindWindow/")

    def press_key(self, key):
        """
        按键
        """
        params = {
            "key": key
        }
        requests.post(url=f"http://localhost:{self.port}/PressKey/", params=params)

    def left_click(self, x, y):
        params = {
            "x": x,
            "y": y,
            "right": False
        }
        res = requests.post(url=f"http://localhost:{self.port}/Click/", params=params)
        print(res)

    def right_click(self, x, y):
        params = {
            "x": x,
            "y": y,
            "right": True
        }
        res = requests.post(url=f"http://localhost:{self.port}/Click/", params=params)
        print(res)

    def drag_to(self, from_point, to_point, button="left"):
        params = {
            "x1": from_point[0],
            "y1": from_point[1],
            "x2": to_point[0],
            "y2": to_point[1],
            "button": button
        }
        res = requests.post(url=f"http://localhost:{self.port}/Drag/", params=params)
        print(res)


class DM:
    def __init__(self, key="mmdlljafd8c5cbf193e99306e9eb61ceb5bd44", add_key="ewdSEKFZP"):
        self.dm = None
        self.key = key
        self.add_key = add_key
        self.bind_mode: list | None = None
        self.load_dm()
        self.is_bind = False

    def load_dm(self):
        try:
            self.dm = win32com.client.Dispatch('dm.dmsoft')
            print('本机系统中已经安装大漠插件，版本为:', self.dm.ver())
        except:
            print('本机并未安装大漠，正在免注册调用')
            dms = ctypes.windll.LoadLibrary('C://Users/DmReg.dll')
            location_dmreg = 'C://Users/dmyk/dm.dll'
            dms.SetDllPathW(location_dmreg, 0)
            self.dm = win32com.client.Dispatch('dm.dmsoft')

            # dm = CreateObject('dm.dmsoft')
            print('免注册调用成功 版本号为:', self.dm.Ver())

            res = self.dm.Reg(self.key, self.add_key)
            if res == 1:
                print('大漠注册成功！')
            elif res == -1:
                logger.error("大漠插件无法连接网络")
            elif res == -2:
                logger.error("进程没有以管理员方式运行. (出现在win7 win8 vista 2008.建议关闭uac)")
            elif res == 2:
                logger.error("大漠余额不足")
            elif res == 3:
                logger.error("绑定了本机器，但是账户余额不足50元.")
            elif res == 4:
                logger.error("注册码错误")
            elif res == 5:
                logger.error("你的机器或者IP在黑名单列表中或者不在白名单列表中")
            elif res == 6:
                logger.error(
                    "非法使用插件. 一般出现在定制插件时，使用了和绑定的用户名不同的注册码.  也有可能是系统的语言设置不是中文简体,也可能有这个错误.")
            elif res == 7:
                logger.error(
                    "你的帐号因为非法使用被封禁. （如果是在虚拟机中使用插件，必须使用Reg或者RegEx，不能使用RegNoMac或者RegExNoMac,否则可能会造成封号，或者封禁机器）")
            elif res == 8:
                logger.error("ver_info不在你设置的附加白名单中.")

    def DmGuard(self, enable, type_="memory4"):
        res = self.dm.DmGuard(1, "memory4")
        if res == 1:
            print("大漠防护盾开启成功")
        else:
            logger.warning(f"大漠防护盾开启失败，错误码：{res}")

    def set_bindmode(self, display="dx.graphic.opengl",
                     mouse="dx.mouse.position.lock.api|dx.mouse.position.lock.message|dx.mouse.clip.lock.api|"
                           "dx.mouse.input.lock.api|dx.mouse.state.api|dx.mouse.api|dx.mouse.cursor",
                     keypad="windows",
                     public="dx.public.active.api|dx.public.active.message|dx.public.hide.dll|dx.public.active.api2|dx."
                            "public.anti.api|dx.public.km.protect|dx.public.inject.super|dx.public.memory|"
                            "dx.public.inject.c",
                     mode=11):
        """
        设置大漠的窗口绑定方式，参考BindWindow参数说明
        """
        self.bind_mode = [display, mouse, keypad, public, mode]

    def bind_window(self, hwnd):
        if self.dm is None:
            self.load_dm()
        res = self.dm.BindWindowEx(hwnd, self.bind_mode[0], self.bind_mode[1], self.bind_mode[2], self.bind_mode[3],
                                   self.bind_mode[4])
        if res == 1:
            self.is_bind = True
        else:
            code = self.dm.GetLastError()
            logger.error(f"大漠绑定窗口失败，错误码：{code}")

    def press_key(self, key):
        """
        按键
        """
        self.dm.KeyPress(key)

    def left_click(self, x, y):
        self.dm.LeftClick(x, y)

    def right_click(self, x, y):
        self.dm.RightClick(x, y)


class ThreadXMLRpcServer(ThreadingMixIn, SimpleXMLRPCServer):
    pass


def click(x, y, right=False):
    if right:
        dm.right_click(x, y)
    else:
        dm.left_click(x, y)


def start_rpc_server():
    server = ThreadXMLRpcServer(("localhost", 8888))
    server.register_function(click, "click")
    server.serve_forever()


def start_dm_server(window_hwnd, display, mouse, keypad, public, mode):
    global dm
    dm = DM()
    dm.set_bindmode(display, mouse, keypad, public, mode)
    dm.bind_window(window_hwnd)
    start_rpc_server()


if __name__ == "__main__":
    dm = DM()



import os
import socket
import base64
import hashlib
import json
import traceback
from pyecharts import options
from pyecharts.charts import Bar
from yangke.base import start_threads
from yangke.common.config import logger
from yangke.common.fileOperate import read_lines, write_lines

connections_available = []
stop_serve = False
ready = False


class DataPackage:
    def __init__(self, cmd: str, option=None, series=None, args: dict | None = None):
        """
        发送给html网页端的数据包结构，pyecharts使用，用于其他数据包发送时，需要通过args传递参数。

        :param cmd: 命令，js端根据该命令执行相应的处理。
        :param option: echarts的option对象
        :param series: echarts的series对象
        :param args: 发送给前端的数据对象，是一个字典。如果命令不需要参数，可以不指定该参数。
        """
        self.cmd = cmd
        self.option = option
        self.series = series
        self.args = args

    def to_msg(self):
        """
        将当前数据包转换为json字符串
        """
        if self.args is None:
            self.args = {"cmd": self.cmd, "option": self.option, "series": self.series}
        else:
            self.args.update({"cmd": self.cmd, "option": self.option, "series": self.series})
        return self.args

    def send(self):
        send_msg(self.to_msg())


def get_headers(data):
    """
    解析浏览器端发送到websocket的数据头
    """
    header_dict = {}
    data = data.decode(encoding='utf-8')

    header, body = data.split("\r\n\r\n", 1)
    header_list = header.split("\r\n")
    for i in range(0, len(header_list)):
        if i == 0:
            if len(header_list[i].split(' ')) == 3:
                header_dict['method'], header_dict['url'], header_dict['protocol'] = header_list[i].split(' ')
        else:
            k, v = header_list[i].split(":", 1)
            header_dict[k] = v.strip()
    return header_dict


def send_msg(msg):
    """
    WebSocket服务端向客户端发送消息。即python端向html网页发送数据。
    示例1：
    send_msg("你好")
    示例2：
    from pyecharts.charts import Bar
    c = (
    Bar()
    .add_xaxis(["衬衫", "羊毛衫", "雪纺衫", "裤子", "高跟鞋", "袜子"])
    .add_yaxis("商家A", [5, 20, 36, 10, 75, 90])
    .add_yaxis("商家B", [15, 25, 16, 55, 48, 8])
    .set_global_opts(title_opts=options.TitleOpts(title="Bar-基本示例", subtitle="我是副标题"))
    )
    message = c.dump_options_with_quotes()
    send_msg(message)

    :param msg: 向客户端发送的数据，可以是字符串或字典
    :return:
    """
    import struct
    if isinstance(msg, str):
        print(f"发送消息：{msg}")
        msg = msg.encode('utf8')
    elif isinstance(msg, dict) or isinstance(msg, list):
        print(f"发送消息：dict object")
        msg = json.dumps(msg)
        msg = msg.encode('utf8')
    # 不知道以下编码规则是啥，尝试删除看还能不能正常发送数据
    token = b"\x81"

    length = len(msg)
    if length < 126:
        token += struct.pack("B", length)
    elif length <= 0xFFFF:
        token += struct.pack("!BH", 126, length)
    else:
        token += struct.pack("!BQ", 127, length)

    msg = token + msg
    for conn in connections_available:
        try:
            conn.send(msg)
        except ConnectionAbortedError:  # 强行关闭前端网页时，可能出现连接实际已经不存在，但后台还有记录
            pass
    return True


def start_ws_serve(port=8017, ip='127.0.0.1', callback=None):
    """
    在计算机某个端口上建立长连接监听

    :param port:
    :param ip:
    :param callback: 服务端接收到前端画面返回的信息后，调用的函数
    :return:
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((ip, port))
    sock.listen(5)  # 允许的连接数量
    global connections_available
    print(f"服务建立：{ip}:{port}")
    global stop_serve, ready
    stop_serve = False
    ready = True
    while True:
        conn, address = sock.accept()
        data = conn.recv(1024)  # 阻塞式监听
        headers = get_headers(data)  # 网页端打开并初次建立ws连接时，会发送header数据
        response_tpl = "HTTP/1.1 101 Switching Protocols\r\n" \
                       "Upgrade:websocket\r\n" \
                       "Connection:Upgrade\r\n" \
                       "Sec-WebSocket-Accept:%s\r\n" \
                       "WebSocket-Location:ws://%s%s\r\n\r\n"

        value = headers['Sec-WebSocket-Key'] + '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
        ac = base64.b64encode(hashlib.sha1(value.encode('utf-8')).digest())
        response_str = response_tpl % (ac.decode('utf-8'), headers['Host'], headers['url'])
        conn.send(bytes(response_str, encoding='utf-8'))
        start_threads(on_message, (conn, callback))  # 启动新线程处理该连接，相当于调用on_message(conn, callback)
        print(f"接收到新的连接：{response_str=}, {conn=}")
        connections_available.append(conn)  # 记录已建立的连接，以方便在退出时关闭这些连接
        if stop_serve:  # 当外界将stop_serve设置为True时，则长连接退出
            break
    for conn in connections_available:
        conn.close()
    sock.close()


def decode_bytes(ws_bytes):
    """
    解码前端长连接返回的数据，解码原理未知，反正不是UTF8

    :param ws_bytes:
    :return:
    """
    if len(ws_bytes) == 0:
        return ""
    payload_len = ws_bytes[1] & 127
    if payload_len == 126:
        extend_payload_len = ws_bytes[2:4]
        mask = ws_bytes[4:8]
        decoded = ws_bytes[8:]
    elif payload_len == 127:
        extend_payload_len = ws_bytes[2:10]
        mask = ws_bytes[10:14]
        decoded = ws_bytes[14:]
    else:
        extend_payload_len = None
        mask = ws_bytes[2:6]
        decoded = ws_bytes[6:]

    bytes_list = bytearray()
    for i in range(len(decoded)):
        chunk = decoded[i] ^ mask[i % 4]
        bytes_list.append(chunk)
    try:
        message = bytes_list.decode("utf8")
        return message
    except UnicodeDecodeError:
        if bytes_list == b"\x03\xe9":
            return "close"
        else:
            logger.warning(f"解码字节流{bytes_list}时出错！")


def on_message(conn: socket.socket, callback=None):
    """
    该线程一直监听端口，接收前端发回的信息，并处理

    :param conn:
    :param callback: 后台接收到消息后，执行的回调函数
    :return:
    """
    while True:
        try:
            info = conn.recv(8096)
        except Exception as e:
            logger.warning("发生异常，但服务仍继续")
            break
        info = decode_bytes(info)
        if info == "close":
            print(f"conn-{conn.getpeername()[1]}连接已关闭！")
            if conn in connections_available:
                connections_available.remove(conn)
            conn.close()
            break
        elif info != "":
            print(f"服务器接收到conn-{conn.getpeername()[1]}的消息：{info}")
            if info == "test":
                c = (
                    Bar()
                    .add_xaxis(["衬衫", "羊毛衫", "雪纺衫", "裤子", "高跟鞋", "袜子"])
                    .add_yaxis("商家A", [5, 20, 36, 10, 75, 90])
                    .add_yaxis("商家B", [15, 25, 16, 55, 48, 8])
                    .set_global_opts(title_opts=options.TitleOpts(title="Bar-基本示例", subtitle="我是副标题"))
                )
                echarts_option = c.dump_options_with_quotes()  # pyecharts将图表设置转换为网页端可以理解的字符串，类似于字典
                data_package = DataPackage(cmd="initChart", option=echarts_option)
                data_package.send()  # 发送数据包
            if type(callback).__name__ == "method" or type(callback).__name__ == "function":
                callback(info)


def stop_ws_serve():
    global ready
    global stop_serve
    stop_serve = True
    ready = False


def is_ready():
    return ready


if __name__ == "__main__":
    start_ws_serve()



import numpy as np


def selection_tournament_faster(self, tourn_size=3):
    '''
    Select the best individual among *tournsize* randomly chosen
    Same with `selection_tournament` but much faster using numpy
    individuals,
    :param self:
    :param tourn_size:
    :return:
    '''
    aspirants_idx = np.random.randint(self.size_pop, size=(self.size_pop, tourn_size))
    aspirants_values = self.FitV[aspirants_idx]
    winner = aspirants_values.argmax(axis=1)  # winner index in every team
    sel_index = [aspirants_idx[i, j] for i, j in enumerate(winner)]
    self.Chrom = self.Chrom[sel_index, :]
    return self.Chrom



import numpy as np
import dataset.data_reader as dr
import paddle.fluid as fluid
import paddle
import globalVar as gv


def prediction1(datafile):
    """
    根据过去30天的股票数据，预测下一天收盘时的股价涨幅

    :param datafile: 股票数据存储的文件
    :return:
    """
    epochs = 10
    BATCH_SIZE = 10

    # 生成数据集
    # days, cols, labels意义参见dr.get_data_days_for_change()方法
    # 这里生成的是用于分类预测的数据集，按涨幅是否大于1%将标签分为两类
    days, cols, labels = 30, ['open', 'high', 'low', 'close', 'volume'], [1]
    generator, _, _ = dr.get_data_days_generator(datafile, days, cols, labels, normalization='min-max')

    # x, y = dr.get_data_days_list(datafile, days, cols, labels)

    # 定义数据的形状
    stocks_info = fluid.layers.data(name='stocks_info', shape=[days, len(cols)], dtype='float32')
    label = fluid.layers.data(name='label', shape=[1], dtype='int64')

    # 定义网络结构
    label_pred = fluid.layers.fc(input=stocks_info, size=len(labels) + 1, act='softmax')
    loss = fluid.layers.cross_entropy(input=label_pred, label=label)
    avg_loss = fluid.layers.mean(loss)
    acc = fluid.layers.accuracy(input=label_pred, label=label)

    # 定义优化器
    optimizer = fluid.optimizer.Adam(learning_rate=0.001)
    optimizer.minimize(avg_loss)

    # 定义调试器
    place = fluid.CPUPlace()
    exe = fluid.Executor(place)
    exe.打开技术报告页面(fluid.default_startup_program())  # 初始化调试器
    main_program = fluid.default_main_program()

    # 定义数据的generator
    train_reader = paddle.batch(generator,
                                batch_size=BATCH_SIZE)  # 一次拿到BATCH_SIZE个数据
    feeder = fluid.DataFeeder(feed_list=[stocks_info, label], place=place)

    epochs_list = [epoch_id for epoch_id in range(epochs)]
    for epoch_id in epochs_list:
        for step, data in enumerate(train_reader()):
            metrics = exe.打开技术报告页面(
                main_program, feed=feeder.feed(data), fetch_list=[avg_loss, acc]
            )
            step += 1
            if step % 10 == 0:
                gv.logger.debug(
                    "Epoch: {}; Step: {}; Cost: {}; Accuracy: {}".format(epoch_id, step, metrics[0], metrics[1]))
        gv.logger.debug("avg_loss is {}, accuracy is {}".format(metrics[0], metrics[1]))
